__author__ = 'Joe Linn'

import json
import pylastica.index


class Action(object):
    OP_TYPE_CREATE = 'create'
    OP_TYPE_INDEX = 'index'
    OP_TYPE_DELETE = 'delete'
    OP_TYPE_UPDATE = 'update'

    def __init__(self, op_type=OP_TYPE_INDEX, metadata=None, source=None):
        """

        @param op_type: see OP_TYPE_* class properties for options
        @type op_type: str
        @param metadata:
        @type metadata: dict
        @param source:
        @type source: dict
        """
        self.op_type = op_type
        self.metadata = metadata
        self.source = source

    @property
    def op_type(self):
        """

        @return:
        @rtype: str
        """
        return self._op_type

    @op_type.setter
    def op_type(self, op_type):
        """

        @param op_type: see OP_TYPE_* class properties for options
        @type op_type: str
        """
        self._op_type = op_type

    @property
    def metadata(self):
        """

        @return:
        @rtype: dict
        """
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        """

        @param metadata:
        @type metadata: dict
        """
        if metadata is None:
            metadata = {}
        self._metadata = metadata

    @property
    def action_metadata(self):
        """

        @return:
        @rtype: dict
        """
        return {self.op_type: self.metadata}

    @property
    def source(self):
        """

        @return:
        @rtype: dict
        """
        return self._source

    @source.setter
    def source(self, source):
        """

        @param source:
        @type source: dict
        """
        self._source = source

    def has_source(self):
        """

        @return:
        @rtype: bool
        """
        return self._source is not None

    def set_index(self, index):
        """

        @param index:
        @type index: str or pylastica.index.Index
        @return:
        @rtype: self
        """
        if isinstance(index, pylastica.index.Index):
            index = index.name
        self._metadata['_index'] = index
        return self

    def set_type(self, doc_type):
        """

        @param doc_type:
        @type doc_type: str or pylastica.doc_type.DocType
        @return:
        @rtype: self
        """
        if isinstance(doc_type, pylastica.doc_type.DocType):
            self.set_index(doc_type.index.name)
            doc_type = doc_type.name
        self._metadata['_type'] = type
        return self

    def set_id(self, doc_id):
        """

        @param doc_id:
        @type doc_id: str
        @return:
        @rtype: self
        """
        self._metadata['_id'] = doc_id
        return self

    def to_list(self):
        """

        @return:
        @rtype: list
        """
        data = [self.action_metadata]
        if self.has_source():
            data.append(self.source)
        return data

    def to_string(self):
        """

        @return:
        @rtype: str
        """
        from pylastica.bulk import Bulk
        string = json.dumps(self.action_metadata) + Bulk.DELIMITER
        if self.has_source():
            source = self.source
            if isinstance(source, str):
                string += source
            else:
                string += json.dumps(source)
            string += Bulk.DELIMITER
        return string

    def __str__(self):
        """

        @return:
        @rtype: str
        """
        return self.to_string()

    @staticmethod
    def is_valid_op_type(op_type):
        """
        Determines whether or not the given op type is a valid bulk operation
        @param op_type:
        @type op_type: str
        @return:
        @rtype: bool
        """
        valid = [
            Action.OP_TYPE_INDEX,
            Action.OP_TYPE_CREATE,
            Action.OP_TYPE_DELETE,
            Action.OP_TYPE_UPDATE
        ]
        return op_type in valid




__author__ = 'Joe Linn'

import pylastica.aggregation.abstractaggregation as abstract
import pylastica.filter.abstractfilter as abstractfilter


class Filter(abstract.AbstractAggregation):
    def set_filter(self, filter):
        """
        Set the filter for this aggregation
        @param filter: the filter to use for this aggregation
        @type filter: pylastica.filter.abstractfilter.AbstractFilter
        @return:
        @rtype: Filter
        """
        if not isinstance(filter, abstractfilter.AbstractFilter):
            raise TypeError("filter must be an instance of an implementation of AbstractFilter: %r" % filter)
        return self.set_param("filter", filter.to_dict())

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        return {
            "filter": self.get_param("filter"),
            "aggs": self._aggs
        }


__author__ = 'Joe Linn'

from .abstract import AbstractRiver


class RabbitMQ(AbstractRiver):
    """
    @see: https://github.com/elasticsearch/elasticsearch-river-rabbitmq/blob/master/README.md
    """

    river_type = 'rabbitmq'

    EXCHANGE_DIRECT = 'direct'
    EXCHANGE_FANOUT = 'fanout'

    def __init__(self, client, name, index=None, doc_type=None, bulk_size=100, bulk_timeout='10ms',
                 host='localhost', port=5672, user='guest', password='guest'):
        """
        @param client:
        @type client: pylastica.client.Client
        @param name: name of the river
        @type name: str
        @param index: default index for this river
        @type index: str or pylastica.index.Index
        @param doc_type: default document type for this river
        @type doc_type: str or pylastica.doc_type.DocType
        @param bulk_size: bulk size
        @type bulk_size: int
        @param bulk_timeout: "10ms", for example
        @type bulk_timeout: str
        @param host: RabbitMQ hostname or ip address
        @type host: str
        @param port: RabbitMQ port
        @type port: int
        @param user: RabbitMQ user
        @type user: str
        @param password: RabbitMQ password
        @type password: str
        """
        super(RabbitMQ, self).__init__(client, name, index, doc_type, bulk_size, bulk_timeout)
        self.set_host(host).set_port(port).set_user(user).set_password(password)

    def set_host(self, host):
        """
        Set RabbitMQ host
        @param host: hostname or ip address
        @type host: str
        @return:
        @rtype: self
        """
        return self.set_param('host', host)

    def set_port(self, port):
        """
        Set RabbitMQ port
        @param port: port number
        @type port: int
        @return:
        @rtype: self
        """
        return self.set_param('port', port)

    def set_user(self, user):
        """
        Set RabbitMQ user
        @param user: username
        @type user: str
        @return:
        @rtype: self
        """
        return self.set_param('user', user)

    def set_password(self, password):
        """
        Set RabbitMQ password
        @param password:
        @type password: str
        @return:
        @rtype: self
        """
        return self.set_param('pass', password)

    def set_vhost(self, vhost='/'):
        """
        Set the RabbitMQ vhost
        @param vhost: defaults to /
        @type vhost: str
        @return:
        @rtype: self
        """
        return self.set_param('vhost', vhost)

    def set_queue(self, queue='elasticsearch'):
        """
        Set the RabbitMQ queue name
        @param queue:
        @type queue: str
        @return:
        @rtype: self
        """
        return self.set_param('queue', queue)

    def set_exchange(self, exchange='elasticsearch'):
        """
        Set the RabbitMQ exchange
        @param exchange:
        @type exchange:
        @return:
        @rtype:
        """
        return self.set_param('exchange', exchange)

    def set_routing_key(self, routing_key='elasticsearch'):
        """
        Set the routing key
        @param routing_key:
        @type routing_key: str
        @return:
        @rtype: self
        """
        return self.set_param('routing_key', routing_key)

    def set_exchange_declare(self, declare=True):
        """
        Enable / disable exchange declaration
        @param declare: defaults to True
        @type declare: bool
        @return:
        @rtype: self
        """
        return self.set_param('exchange_declare', bool(declare))

    def set_exchange_type(self, exchange_type='direct'):
        """
        Set exchange type
        @param exchange_type: see EXCHANGE_* class properties for options
        @type exchange_type: str
        @return:
        @rtype: self
        """
        return self.set_param('exchange_type', exchange_type)

    def set_exchange_durable(self, durable=True):
        """
        Enable / disable exchange durability
        @param durable: defaults to True
        @type durable: bool
        @return:
        @rtype: self
        """
        return self.set_param('exchange_durable', bool(durable))

    def set_queue_declare(self, declare=True):
        """
        Enable / disable queue declaration
        @param declare: defaults to True
        @type declare: bool
        @return:
        @rtype: self
        """
        return self.set_param('queue_declare', bool(declare))

    def set_queue_durable(self, durable=True):
        """
        Enable / disable queue durability
        @param durable: defaults to True
        @type durable: bool
        @return:
        @rtype: self
        """
        return self.set_param('queue_durable', bool(durable))

    def set_queue_auto_delete(self, auto=False):
        """
        Enable / disable queue auto deletion
        @param auto: defaults to False
        @type auto: bool
        @return:
        @rtype: self
        """
        return self.set_param('queue_auto_delete', bool(auto))



__author__ = 'Joe Linn'

import pylastica.aggregation.abstractaggregation as abstract


class IpRange(abstract.AbstractAggregation):
    def __init__(self, name, field):
        """
        @param name: the name of this aggregation
        @type name: str
        @param field: the name of the document field on which to perform this aggregation
        @type field: str
        """
        super(IpRange, self).__init__(name)
        self.set_field(field)

    def set_field(self, field):
        """
        Set the field for this aggregation
        @param field: the name of the document field on which to perform this aggregation
        @type field: str
        @return:
        @rtype: IpRange
        """
        return self.set_param("field", field)

    def add_range(self, from_value=None, to_value=None):
        """
        Add an ip range to this aggregation
        @param from_value: A valid ipv4 address. Low end of this range, exclusive (greater than)
        @type from_value: str
        @param to_value: A valid ipv4 address. High end of this range, exclusive (less than)
        @type to_value: str
        @return:
        @rtype: IpRange
        """
        if from_value is None and to_value is None:
            raise ValueError("Either from_value or to_value must be set. Both cannot be None.")
        range = {}
        if from_value is not None:
            range['from'] = from_value
        if to_value is not None:
            range['to'] = to_value
        return self.add_param('ranges', range)

    def add_mask_range(self, mask):
        """
        Add an ip range in the form of a CIDR mask
        @param mask: a valid CIDR mask
        @type mask: str
        @return:
        @rtype: IpRange
        """
        return self.add_param("ranges", {"mask": mask})


__author__ = 'Joe Linn'

import pylastica.index


class Percolator(object):

    def __init__(self, index):
        """

        @param index:
        @type index: pylastica.index.Index
        """
        assert isinstance(index, pylastica.index.Index), "index must be an instance of Index: %r" % index
        self._index = index

    def register_query(self, name, query):
        """
        Register a percolator query
        @param name: query name
        @type name: str
        @param query:
        @type query: str or pylastica.query.Query or pylastica.query.AbstractQuery
        @return:
        @rtype: pylastica.response.Response
        """
        path = '%s/.percolator/%s' % (self._index.name, name)
        query = pylastica.query.Query.create(query)
        return self._index.client.request(path, pylastica.request.Request.PUT, query.to_dict())

    def unregister_query(self, name):
        """
        Remove a percolator query
        @param name: name of the query
        @type name: str
        @return:
        @rtype: pylastica.response.Response
        """
        path = '_percolator/%s/%s' % (self._index.name, name)
        return self._index.client.request(path, pylastica.request.Request.DELETE)

    def match_doc(self, doc, query=None):
        """
        Match a document to percolator queries
        @param doc:
        @type doc: pylastica.document.Document
        @param query:
        @type query: str or pylastica.query.Query or pylastica.query.AbstractQuery
        @return:
        @rtype: pylastica.response.Response
        """
        path = "%s/type/_percolate" % self._index.name
        data = {'doc': doc.data}
        if query is not None:
            query = pylastica.query.Query.create(query)
            data['query'] = query.query
        response = self._index.client.request(path, data=data)
        data = response.data
        return data['matches'] if 'matches' in data else []

    @property
    def index(self):
        """

        @return:
        @rtype: pylastica.index.Index
        """
        return self._index



__author__ = 'Joe Linn'

from .indexdocument import IndexDocument


class UpdateDocument(IndexDocument):
    def __init__(self, document):
        """

        @param document:
        @type document: pylastica.document.Document
        """
        super(UpdateDocument, self).__init__(document)
        self._op_type = self.OP_TYPE_UPDATE

    @property
    def document(self):
        """

        @return:
        @rtype: pylastica.document.Document
        """
        return super(UpdateDocument, self).document

    @document.setter
    def document(self, document):
        """
        Set the document for this bulk update operation.
        @param document:
        @type document:  pylastica.document.Document
        """
        super(UpdateDocument, self).set_document(document)
        source = {'doc': document.data}
        if document.doc_as_upsert:
            source['doc_as_upsert'] = True
        elif document.has_upsert():
            upsert = document.upsert.data
            if upsert is not None:
                source['upsert'] = upsert
        self.source = source

    @property
    def script(self):
        """

        @return:
        @rtype: pylastica.script.Script
        """
        return super(UpdateDocument, self).script

    @script.setter
    def script(self, script):
        """

        @param script:
        @type script: pylastica.script.Script
        """
        super(UpdateDocument, self).set_script(script)
        source = script.to_dict()
        if script.has_upsert():
            upsert = script.upsert.data
            if upsert is not None:
                source['upsert'] = upsert
        self.source = source

    def _get_metadata_by_script(self, script):
        """

        @param script:
        @type script: pylastica.script.Script
        @return:
        @rtype: dict
        """
        params = [
            'index',
            'type',
            'id',
            'version',
            'version_type',
            'routing',
            'percolate',
            'parent',
            'ttl',
            'timestamp'
        ]
        metadata = script.get_options(params, True)
        return metadata



__author__ = 'Joe Linn'

class Result(object):
    def __init__(self, hit):
        """

        @param hit:
        @type hit: dict
        @return:
        """
        self._hit = hit

    def get_param(self, name):
        """
        Return a param from the result hit dict
        @param name:
        @type name: str
        @return:
        @rtype: dict
        """
        if name in self._hit:
            return self._hit[name]
        return {}

    def get_id(self):
        """
        Returns the hit id
        @return:
        @rtype: str
        """
        return self.get_param('_id')

    def get_type(self):
        """
        Return the type of the result
        @return:
        @rtype: str
        """
        return self.get_param('_type')

    def get_fields(self):
        """
        Returns a list of fields
        @return:
        @rtype: dict
        """
        return self.get_param('fields')

    def get_index(self):
        """
        Returns the index name of the result
        @return:
        @rtype: str
        """
        return self.get_param('_index')

    def get_score(self):
        """
        Return the score of the result
        @return:
        @rtype: float
        """
        return self.get_param('_score')

    def get_hit(self):
        """
        Return the raw hit dict
        @return:
        @rtype: dict
        """
        return self._hit

    def get_version(self):
        """
        Return the version of the hit
        @return:
        @rtype: int
        """
        return self.get_param('_version')

    @property
    def data(self):
        """
        Returns result data
        @return:
        @rtype: dict
        """
        return self.get_data()

    def get_data(self):
        """
        Returns result data
        @return:
        @rtype: dict
        """
        if 'fields' in self._hit and '_source' not in self._hit:
            return self.get_fields()
        else:
            return self.get_source()

    def get_source(self):
        """
        Return the result source
        @return:
        @rtype: dict
        """
        return self.get_param('_source')

    def get_highlights(self):
        """
        Return result data
        @return:
        @rtype: dict
        """
        return self.get_param('highlight')

    def get_explanation(self):
        """
        Return an explanation of how the result's score was computed
        @return:
        @rtype: dict
        """
        return self.get_param('_explanation')

    # def __getattribute__(self, name):
    #     source = self.get_data()
    #     if name in source:
    #         return source[name]
    #     else:
    #         return None





__author__ = 'Joe Linn'

import pylastica.client
import pylastica.request
from . import health
from .settings import *

class Cluster(object):
    def __init__(self, client):
        """

        @param client:
        @type client: pylastica.client.Client
        """
        super(Cluster, self).__init__()
        assert isinstance(client, pylastica.client.Client), "client must be an instance of pylastica.client.Client: %r" % client
        self._client = client
        self._response = None
        self._data = None
        self.refresh()

    def refresh(self):
        """
        Refresh all cluster information
        @return:
        @rtype: self
        """
        self._response = self._client.request('_cluster/state', pylastica.request.Request.GET)
        self._data = self.response.data

    @property
    def pending_tasks(self):
        """
        Get a list of pending cluster tasks
        @return:
        @rtype: pylastica.response.Response
        """
        return self._client.request('_cluster/pending_tasks')

    @property
    def response(self):
        """
        Return the Response object
        @return:
        @rtype: pylastica.response.Response
        """
        return self._response

    @property
    def index_names(self):
        """
        Return a list of index names
        @return:
        @rtype: list of str
        """
        metadata = self._data['metadata']['indices']
        return [key for key in metadata]

    @property
    def state(self):
        """
        Returns the full state of the cluster
        @return:
        @rtype: dict
        @see: http://www.elasticsearch.org/guide/reference/api/admin-cluster-state.html
        """
        return self._data

    @property
    def node_names(self):
        """
        Returns a list of existing node names
        @return:
        @rtype: list of str
        """
        return self.state['routing_nodes']['nodes'].keys()

    @property
    def nodes(self):
        """
        Return all nodes in the cluster
        @return:
        @rtype: list of pylastica.node.Node
        """
        return [pylastica.node.Node(name, self.client) for name in self.node_names]

    @property
    def client(self):
        """
        Return the client object
        @return:
        @rtype: pylastica.client.Client
        """
        return self._client

    def get_info(self, args = None):
        """
        Returns cluster information
        @param args: additional arguments
        @type args: dict
        @return:
        @rtype:
        @see: http://www.elasticsearch.org/guide/reference/api/admin-cluster-nodes-info.html
        """
        return self.client.request('_nodes', query=args)

    @property
    def health(self):
        """
        Return cluster health
        @return:
        @rtype: pylastica.cluster.health.Health
        """
        return pylastica.cluster.health.Health(self.client)

    @property
    def settings(self):
        """
        Return cluster settings
        @return:
        @rtype: pylastica.cluster.settings.Settings
        """
        return pylastica.cluster.settings.Settings(self.client)

    def shutdown(self, delay='1s'):
        """
        Shut down the cluster
        @param delay: (optional) time delay before cluster shutdown (default 1s)
        @type delay: str
        @return:
        @rtype: pylastica.response.Response
        """
        return self.client.request('_shutdown', query={'delay': delay})



__author__ = 'Joe Linn'

import pylastica

class Status(object):
    def __init__(self, index):
        """
        @param index:
        @type index: pylastica.index.Index
        """
        assert isinstance(index, pylastica.index.Index), "index must be an instance of Index: %r" % index
        self._index = index
        self._response = None
        self._data = {}
        self.refresh()

    @property
    def data(self):
        """
        Returns all status info
        @return:
        @rtype: dict
        """
        return self._data

    def get(self, *args):
        """
        Returns the entry in the internal data based on the params.
        @param args:
        @type args: str
        @return:
        @rtype: dict or None
        """
        data = self.data
        data = data['indices'][self.index.name]
        for arg in args:
            if arg in data:
                data = data[arg]
            else:
                return None
        return data

    @property
    def aliases(self):
        """
        Returns all index aliases
        @return:
        @rtype: list
        """
        data = self.index.request('_aliases', pylastica.request.Request.GET).data[self.index.name]
        if 'aliases' not in data:
            raise pylastica.exception.NotFoundException("Alias data for index %s not found." % self.index.name)
        return [name for name in data['aliases']]

    def has_alias(self, alias):
        """
        Checks if the index has the given alias
        @param alias: alias name
        @type alias: str
        @return:
        @rtype: bool
        """
        return alias in self.aliases

    @property
    def settings(self):
        """
        @return: Index settings
        @rtype: dict
        """
        response_data = self.index.request('_settings', pylastica.request.Request.GET).data
        return response_data[self.index.name]['settings']

    @property
    def index(self):
        """
        get the index object
        @return:
        @rtype: pylastica.index.Index
        """
        return self._index

    @property
    def response(self):
        """
        Return the response object
        @return:
        @rtype: pylastica.response.Response
        """
        return self._response

    def refresh(self):
        """
        Reloads all status data for this object
        """
        self._response = self.index.request('_status', pylastica.request.Request.GET)
        self._data = self.response.data



__author__ = 'Joe Linn'

import abc
from pylastica.param import Param


class AbstractSuggestion(Param):
    __metaclass__ = abc.ABCMeta

    def __init__(self, name, field):
        """
        @param name: the name of this suggestion
        @type name: str
        @param field: the field on which to perform this suggestion
        @type field: str
        """
        super(AbstractSuggestion, self).__init__()
        self._name = ''
        self._text = None
        self._name = name
        self.set_field(field)

    def set_text(self, text):
        """
        Suggest text must be set either globally or per suggestion
        @param text: text to use for this suggestion
        @type text: str
        @rtype: self
        """
        self._text = text
        return self

    def set_field(self, field):
        """
        Set the field on which to perform this suggestion operation
        @param field: field name
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param('field', field)

    def set_size(self, size):
        """
        Set the maximum number of suggestions to return
        @param size: a positive integer
        @type size: int
        @return:
        @rtype: self
        """
        return self.set_param("size", int(size))

    def set_shard_size(self, size):
        """
        Set the maximum number of suggestions to be retrieved from each shard
        @param size: a positive integer
        @type size: int
        @return:
        @rtype: self
        """
        return self.set_param("shard_size", int(size))

    @property
    def name(self):
        """
        @return: the name set for this suggestion
        @rtype: str
        """
        return self._name

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        dictionary = super(AbstractSuggestion, self).to_dict()
        if self._text is not None:
            dictionary['text'] = self._text
        return dictionary



__author__ = 'Joe Linn'

import pylastica.exception
from .abstractfilter import AbstractFilter


class Terms(AbstractFilter):
    def __init__(self, key='', terms=None):
        """

        @param key: document field
        @type key: str
        @param terms: list of term values
        @type terms: list of str
        """
        super(Terms, self).__init__()
        if terms is not None:
            self.set_terms(key, terms)

    def set_terms(self, key, terms):
        """
        Set key and terms for the filter
        @param key: document field
        @type key: str
        @param terms: term values
        @type terms: list of str
        @return:
        @rtype: self
        """
        if not isinstance(terms, list):
            raise TypeError("terms must be a list: %r" % terms)
        self._key = key
        self._terms = terms
        return self

    def set_lookup(self, key, doc_type, id, path, index=None):
        """
        Set the lookup path for this terms filter
        @param key: document field
        @type key: str
        @param doc_type: the document type from which to fetch the term values
        @type doc_type: str or pylastica.doc_type.DocType
        @param id: id of the document from which to fetch the term values
        @type id: str
        @param path: the field specified as path from which to fetch the values for the filter
        @type path: str
        @param index: The index from which to fetch the term values. Defaults to the current index.
        @type index: str or pylastica.index.Index
        @return:
        @rtype: self
        """
        self._key = key
        if isinstance(doc_type, pylastica.doc_type.DocType):
            doc_type = doc_type.name
        self._terms = {
            'type': doc_type,
            'id': id,
            'path': path
        }
        if index is not None:
            if isinstance(index, pylastica.index.Index):
                index = index.name
            self._terms['index'] = index
        return self

    def add_term(self, term):
        """
        Add a term to the filter
        @param term:
        @type term: str
        @return:
        @rtype: self
        """
        self._terms.append(term)
        return self

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        if not self._key:
            raise pylastica.exception.InvalidException("Terms key must be set!")
        self._params[self._key] = self._terms
        return {'terms': self._params}




__author__ = 'Joe Linn'

import pylastica.aggregation.abstractaggregation as abstract


class DateRange(abstract.AbstractAggregation):
    def set_field(self, field):
        """
        Set the field for this aggregation
        @param field: the name of the document field on which to perform this aggregation
        @type field: str
        @return:
        @rtype: DateRange
        """
        return self.set_param("field", field)

    def add_range(self, from_value=None, to_value=None):
        """
        Add a date range for this aggregation.
        See documentation for valid date formats.
        @see http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/search-aggregations-bucket-daterange-aggregation.html
        @param from_value: low end of this range, exclusive (greater than)
        @type from_value: str or int
        @param to_value: high end of this range, exclusive (less than)
        @type to_value: str or int
        @return:
        @rtype: DateRange
        """
        if from_value is None and to_value is None:
            raise ValueError("Either from_value or to_value must be set. Both cannot be None.")
        range = {}
        if from_value is not None:
            range['from'] = from_value
        if to_value is not None:
            range['to'] = to_value
        return self.add_param('ranges', range)

    def set_format(self, format):
        """
        Set the format of the date values to be returned
        @param format: see documentation for formatting options
        @type format: str
        @return:
        @rtype: DateRange
        """
        return self.set_param("format", format)


__author__ = 'Joe Linn'

from .abstract import AbstractQuery

class FuzzyLikeThis(AbstractQuery):
    def __init__(self):
        super(FuzzyLikeThis, self).__init__()
        self._fields = []
        self._like_text = ''
        self._ignore_tf = False
        self._max_query_terms = 25
        self._min_similarity = 0.5
        self._prefix_length = 0
        self._boost = 1.0

    def set_fields(self, fields):
        """
        Set fields for this query
        @param fields: field names
        @type fields: list of str
        @return:
        @rtype: self
        """
        self._fields = fields
        return self

    def set_like_text(self, text):
        """
        Set the like_text value
        @param text:
        @type text: str
        @return:
        @rtype: self
        """
        self._like_text = text.strip()
        return self

    def set_ignore_tf(self, ignore=True):
        """
        Set the ignore_tf (term frequency) value
        @param ignore:
        @type ignore: bool
        @return:
        @rtype: self
        """
        self._ignore_tf = bool(ignore)
        return self

    def set_min_similarity(self, value):
        """
        Set the minimum similarity
        @param value:
        @type value: float
        @return:
        @rtype: self
        """
        self._min_similarity = float(value)
        return self

    def set_boost(self, boost):
        """
        Set the boost for this query
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        self._boost = float(boost)
        return self

    def set_prefix_length(self, value):
        """
        Set the prefix length
        @param value:
        @type value: int
        @return:
        @rtype: self
        """
        self._prefix_length = int(value)
        return self

    def set_max_query_terms(self, terms):
        """
        Set the max_query_terms
        @param terms:
        @type terms: int
        @return:
        @rtype: self
        """
        self._max_query_terms = int(terms)
        return self

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        args = {}
        if len(self._fields):
            args['fields'] = self._fields
        if self._boost is not None:
            args['boost'] = self._boost
        if self._like_text:
            args['like_text'] = self._like_text
        args['min_similarity'] = self._min_similarity if self._min_similarity > 0 else 0
        args['prefix_length'] = self._prefix_length
        args['ignore_tf'] = self._ignore_tf
        args['max_query_terms'] = self._max_query_terms
        return {'fuzzy_like_this': args}



__author__ = 'Joe Linn'

from .abstract import AbstractQuery

class Term(AbstractQuery):
    def __init__(self, term=None):
        """

        @param term: optional
        @type term: dict
        """
        super(Term, self).__init__()
        if term is not None:
            self.set_raw_term(term)

    def set_raw_term(self, term):
        """
        Set term.
        @param term:
        @type term: dict
        @return:
        @rtype: self
        """
        self.params = term
        return self

    def set_term(self, key, value, boost=1.0):
        """
        Add a term to the query
        @param key: key to query
        @type key: str
        @param value: value(s) for the query
        @type value: str or list
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        return self.set_raw_term({key: {'value': value, 'boost': boost}})



__author__ = 'Joe Linn'

#import pylastica
import pylastica.response

class ResponseSet(pylastica.response.Response):
    def __init__(self, response, bulk_responses):
        """

        @param response:
        @type response: pylastica.response.Response
        @param bulk_responses:
        @type bulk_responses: list of pylastica.bulk.response.Response
        """
        super(ResponseSet, self).__init__(response.data)
        self._bulk_responses = bulk_responses

    @property
    def bulk_responses(self):
        """

        @return:
        @rtype: list of pylastica.bulk.response.Response
        """
        return self._bulk_responses

    def get_error(self):
        """
        Returns the first found error
        @return:
        @rtype: self
        """
        error = ''
        for response in self.bulk_responses:
            if response.has_error():
                error = response.get_error()
                break
        return error

    @property
    def error(self):
        """

        @return:
        @rtype: str
        """
        return self.get_error()

    def is_ok(self):
        """

        @return:
        @rtype: bool
        """
        ret = True
        for response in self.bulk_responses:
            if not response.is_ok():
                ret = False
                break
        return ret

    def has_error(self):
        """

        @return:
        @rtype: bool
        """
        ret = False
        for response in self.bulk_responses:
            if response.has_error():
                ret = True
                break
        return ret

    def __iter__(self):
        return iter(self.bulk_responses)

    def __len__(self):
        """

        @return:
        @rtype: int
        """
        return len(self.bulk_responses)

    def __getitem__(self, item):
        """

        @param item:
        @type item: int
        @return:
        @rtype: pylastica.bulk.response.Response
        """
        return self.bulk_responses[item]



__author__ = 'Joe Linn'

import abc
import pylastica.param


class AbstractRiver(pylastica.param.Param):
    __metaclass__ = abc.ABCMeta

    river_type = 'river'    #override this in implementing classes

    def __init__(self, client, name, index=None, doc_type=None, bulk_size=100, bulk_timeout='10ms'):
        """
        @param client:
        @type client: pylastica.client.Client
        @param name: name of the river
        @type name: str
        @param index: default index for this river
        @type index: str or pylastica.index.Index
        @param doc_type: default document type for this river
        @type doc_type: str or pylastica.doc_type.DocType
        @param bulk_size: bulk size
        @type bulk_size: int
        @param bulk_timeout: "10ms", for example
        @type bulk_timeout: str
        """
        super(AbstractRiver, self).__init__()
        if self.river_type == 'river':
            raise NotImplementedError("Class property river_type must be overridden in all implementations of AbstractRiver.")
        self._index = {}
        self._params = {}
        self._client = client
        self._name = name
        self.set_index(index).set_doc_type(doc_type).set_bulk_size(bulk_size).set_bulk_timeout(bulk_timeout)

    def set_index(self, index):
        """
        Set the default index for this river
        @param index:
        @type index: str or pylastica.index.Index
        @return:
        @rtype: self
        """
        if isinstance(index, pylastica.index.Index):
            index = index.name
        if isinstance(index, str):
            self._index['name'] = index
        return self

    def set_doc_type(self, doc_type):
        """
        Set the default document type for this river
        @param doc_type:
        @type doc_type: str or pylastica.doc_type.DocType
        @return:
        @rtype: self
        """
        if isinstance(doc_type, pylastica.doc_type.DocType):
            doc_type = doc_type.name
        if isinstance(doc_type, str):
            self._index['type'] = doc_type
        return self

    def set_bulk_size(self, bulk_size):
        """
        Set the bulk size for this river
        @param bulk_size:
        @type bulk_size: int
        @return:
        @rtype: self
        """
        self._index['bulk_size'] = int(bulk_size)
        return self

    def set_bulk_timeout(self, bulk_timeout):
        """
        Set bulk timeout for this river
        @param bulk_timeout: "10ms", for example
        @type bulk_timeout: str
        @return:
        @rtype: self
        """
        self._index['bulk_timeout'] = bulk_timeout
        return self

    def create(self):
        """

        @return:
        @rtype: pylastica.response.Response
        """
        return self._client.request('/_river/%s/_meta' % self._name, pylastica.request.Request.PUT, self.to_dict())

    def delete(self):
        """

        @return:
        @rtype: pylastica.response.Response
        """
        return self._client.request('/_river/%s/' % self._name, pylastica.request.Request.DELETE)

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        return {
            'type': self.river_type,
            self.river_type: self._params,
            'index': self._index
        }



__author__ = 'Joe Linn'

from .abstractfacet import AbstractFacet
import pylastica.script

class Statistical(AbstractFacet):
    def set_field(self, field):
        """
        Set the field for the stats facet
        @param field:
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param('field', field)

    def set_fields(self, fields):
        """
        Set multiple fields for the facet
        @param fields: list of field names
        @type fields: list of str
        @return:
        @rtype: self
        """
        return self.set_param('fields', fields)

    def set_script(self, script):
        """
        Set the script to calculate stats
        @param script:
        @type script: pylastica.script.Script
        @return:
        @rtype: self
        """
        if not isinstance(script, pylastica.script.Script):
            raise TypeError("script must be of type Script: %r" % script)
        return self.set_param('script', script.script).set_param('params', script.params)

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        self._set_facet_param('statistical', self._params)
        return super(Statistical, self).to_dict()





__author__ = 'Joe Linn'

from .abstractdocument import AbstractDocument


class IndexDocument(AbstractDocument):
    def __init__(self, document):
        """
        @param document:
        @type document: pylastica.document.Document
        """
        super(IndexDocument, self).__init__(document)
        self._op_type = self.OP_TYPE_INDEX

    @property
    def document(self):
        """

        @return:
        @rtype: pylastica.document.Document
        """
        return super(IndexDocument, self).document

    @document.setter
    def document(self, document):
        """
        @param document:
        @type document:  pylastica.document.Document
        """
        super(IndexDocument, self).set_document(document)
        self.source = document.data

    def _get_metadata_by_document(self, document):
        """

        @param document:
        @type document: pylastica.document.Document
        @return:
        @rtype: dict
        """
        params = [
            'index',
            'type',
            'id',
            'version',
            'version_type',
            'routing',
            'percolate',
            'parent',
            'ttl',
            'timestamp'
        ]
        metadata = document.get_options(params, True)
        return metadata



__author__ = 'Joe Linn'

from .abstract import AbstractQuery


class Common(AbstractQuery):
    OPERATOR_AND = 'and'
    OPERATOR_OR = 'or'

    def __init__(self, field, query, cutoff_frequency):
        """
        @param field: The field on which to query
        @type field: str
        @param query: The query string
        @type query: str
        @param cutoff_frequency: percentage in decimal form (.001 == 0.1%)
        @type cutoff_frequency: float
        """
        super(Common, self).__init__()
        self._query_params = {}
        self.set_field(field)
        self.set_query(query)
        self.set_cutoff_frequency(cutoff_frequency)

    def set_field(self, field):
        """
        Set the field on which to query
        @param field: The field on which to query
        @type field: str
        @return:
        @rtype: self
        """
        self._field = field
        return self

    def set_query(self, query):
        """
        Set the query string for this query
        @param query: the query string
        @type query: str
        @return:
        @rtype: self
        """
        return self.set_query_param('query', query)

    def set_cutoff_frequency(self, frequency):
        """
        Set the frequency below which terms will be put in the low frequency group
        @param frequency: percentage in decimal form (.001 == 0.1%)
        @type frequency: float
        @return:
        @rtype: self
        """
        return self.set_query_param('cutoff_frequency', float(frequency))

    def set_low_frequency_operator(self, operator):
        """
        Set the logic operator for low frequency terms
        @param operator: See OPERATOR_* attributes for options.
        @type operator: str
        @return:
        @rtype: self
        """
        return self.set_query_param('low_freq_operator', operator)

    def set_high_frequency_operator(self, operator):
        """
        Set the logic operator for high frequency terms
        @param operator: See OPERATOR_* attributes for options.
        @type operator: str
        @return:
        @rtype: self
        """
        return self.set_query_param('high_freq_operator', operator)

    def set_minimum_should_match(self, minimum):
        """
        Specify the minimum number of low frequency terms which must be present
        @param minimum: minimum number of low frequency terms which must be present
        @type minimum: int
        @return:
        @rtype: self
        """
        return self.set_query_param('minimum_should_match', minimum)

    def set_boost(self, boost):
        """
        Set the boost for this query
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        return self.set_query_param('boost', float(boost))

    def set_analyzer(self, analyzer):
        """
        Set the analyzer for this query
        @param analyzer:
        @type analyzer: str
        @return:
        @rtype: self
        """
        return self.set_query_param('analyzer', analyzer)

    def set_disable_coord(self, disable=True):
        """
        Enable / disable computation of score factor based on the fraction of all query terms contained in the document
        @param disable: coord is enabled by default
        @type disable: bool
        @return:
        @rtype: self
        """
        return self.set_query_param('disable_coord', bool(disable))
    
    def set_query_param(self, key, value):
        """
        Set a parameter which will go in the body of this query
        @param key: parameter key
        @type key: str
        @param value: parameter value
        @type value: mixed
        @return:
        @rtype: self
        """
        self._query_params[key] = value
        return self

    def to_dict(self):
        """
        @return:
        @rtype: dict
        """
        self.set_param(self._field, self._query_params)
        return super(Common, self).to_dict()






__author__ = 'Joe Linn'

from .abstractfilter import AbstractFilter
import pylastica.query.abstract


class Nested(AbstractFilter):
    SCORE_MODE_AVG = 'avg'
    SCORE_MODE_TOTAL = 'total'
    SCORE_MODE_MAX = 'max'
    SCORE_MODE_NONE = 'none'

    def set_path(self, path):
        """
        Set the nested document path
        @param path: document path
        @type path: str
        @return:
        @rtype: self
        """
        return self.set_param('path', path)

    def set_query(self, query):
        """
        Set the nested query
        @param query: query object
        @type query: pylastica.query.abstract.AbstractQuery
        @return:
        @rtype: self
        """
        if not isinstance(query, pylastica.query.abstract.AbstractQuery):
            raise TypeError("query must be of type AbstractQuery: %r" % query)
        return self.set_param('query', query.to_dict())

    def set_score_mode(self, mode):
        """
        Set the score mode
        @param mode: see SCORE_MODE_* class properties for options
        @type mode: str
        @return:
        @rtype: self
        """
        return self.set_param('score_mode', mode)



__author__ = 'Joe Linn'

#import pylastica
import pylastica.client
import pylastica.query

class Search(object):
    OPTION_SEARCH_TYPE = 'search_type'
    OPTION_ROUTING = 'routing'
    OPTION_PREFERENCE = 'preference'
    OPTION_VERSION = 'version'
    OPTION_TIMEOUT = 'timeout'
    OPTION_FROM = 'from'
    OPTION_SIZE = 'size'
    OPTION_SCROLL = 'scroll'
    OPTION_SCROLL_ID = 'scroll_id'

    SEARCH_TYPE_COUNT = 'count'
    SEARCH_TYPE_SCAN = 'scan'
    SEARCH_TYPE_DFS_QUERY_THEN_FETCH = 'dfs_query_then_fetch'
    SEARCH_TYPE_DFS_QUERY_AND_FETCH = 'dfs_query_and_fetch'
    SEARCH_TYPE_QUERY_THEN_FETCH = 'query_then_fetch'
    SEARCH_TYPE_QUERY_AND_FETCH = 'query_and_fetch'

    def __init__(self, client):
        """
        @param client:
        @type client: pylastica.client.Client
        """
        assert isinstance(client, pylastica.client.Client), "client must be an instance of Client: %r" % client
        self._client = client
        self._indices = []
        self._types = []
        self._query = None
        self._options = {}

    def add_index(self, index):
        """
        Add an index to the list
        @param index:
        @type index: str or pylastica.index.Index
        @return:
        @rtype: self
        """
        if isinstance(index, pylastica.index.Index):
            index = index.name
        assert isinstance(index, str), "Invalid param type: %r" % index
        self._indices.append(index)
        return self

    def add_indices(self, indices):
        """
        Add multiple indices to the internal list
        @param indices:
        @type indices: list of str or list of pylastica.index.Index
        @return:
        @rtype: self
        """
        for index in indices:
            self.add_index(index)
        return self

    def add_type(self, doc_type):
        """
        Add a type to the current search
        @param doc_type:
        @type doc_type: str or pylastica.doc_type.DocType
        @return:
        @rtype: self
        """
        if isinstance(doc_type, pylastica.doc_type.DocType):
            doc_type = doc_type.name
        assert isinstance(doc_type, str), "Invalid param type: %r" % doc_type
        self._types.append(doc_type)
        return self

    def add_types(self, doc_types):
        """
        Add multiple document types to the search
        @param doc_types:
        @type doc_types: list of str or list of pylastica.doc_type.DocType
        @return:
        @rtype: self
        """
        for doc_type in doc_types:
            self.add_type(doc_type)
        return self

    def set_query(self, query):
        """
        @param query:
        @type query: str or dict or pylastica.query.Query
        @return:
        @rtype: self
        """
        self._query = pylastica.query.Query.create(query)
        return self

    def set_option(self, key, value):
        """

        @param key:
        @type key: str
        @param value:
        @type value: mixed
        @return:
        @rtype: self
        """
        self._validate_option(key)
        self._options[key] = value
        return self

    def set_options(self, options):
        """
        Set all options. Clears existing options.
        @param options:
        @type options: dict
        @return:
        @rtype: self
        """
        self.clear_options()
        for key, value in options.iteritems():
            self.set_option(key, value)
        return self

    def clear_options(self):
        """

        @return:
        @rtype: self
        """
        self._options = {}
        return self

    def add_option(self, key, value):
        """
        Add an option
        @param key:
        @type key: str
        @param value:
        @type value: mixed
        @return:
        @rtype: self
        """
        self._validate_option(key)
        if key not in self._options:
            self._options[key] = []
        self._options[key].append(value)
        return self

    def has_option(self, key):
        """

        @param key:
        @type key: str
        @return:
        @rtype: bool
        """
        return key in self._options

    def get_option(self, key):
        """

        @param key:
        @type key: str
        @return:
        @rtype: mixed
        """
        if not self.has_option(key):
            raise pylastica.exception.InvalidException("Option %s does not exist." % key)
        return self._options[key]

    @property
    def options(self):
        """

        @return:
        @rtype: dict
        """
        return self._options

    @options.setter
    def options(self, options):
        """
        Set all options. Clears existing options.
        @param options:
        @type options: dict
        """
        self.set_options(options)

    @options.deleter
    def options(self):
        self.clear_options()

    def _validate_option(self, key):
        """

        @param key:
        @type key: key
        @return:
        @rtype: bool
        """
        valid_options = [
            self.OPTION_SEARCH_TYPE,
            self.OPTION_ROUTING,
            self.OPTION_PREFERENCE,
            self.OPTION_PREFERENCE,
            self.OPTION_VERSION,
            self.OPTION_TIMEOUT,
            self.OPTION_FROM,
            self.OPTION_SIZE,
            self.OPTION_SCROLL,
            self.OPTION_SCROLL_ID
        ]
        if key in valid_options:
            return True
        raise pylastica.exception.InvalidException("Invalid option: %s." % key)

    @property
    def client(self):
        """

        @return:
        @rtype: pylastica.client.Client
        """
        return self._client

    @property
    def indices(self):
        """

        @return:
        @rtype: list of str
        """
        return self._indices

    def has_indices(self):
        """

        @return:
        @rtype: bool
        """
        return len(self._indices) > 0

    def has_index(self, index):
        """

        @param index: index object or index name
        @type index: str or pylastica.index.Index
        @return:
        @rtype: bool
        """
        if isinstance(index, pylastica.index.Index):
            index = index.name
        return index in self._indices

    @property
    def doc_types(self):
        """
        Get a list of document types
        @return:
        @rtype: list of str
        """
        return self._types

    def has_doc_types(self):
        """

        @return:
        @rtype: bool
        """
        return len(self._types) > 0

    def has_doc_type(self, doc_type):
        """

        @param doc_type:
        @type doc_type: str or pylastica.doc_type.DocType
        @return:
        @rtype: bool
        """
        if isinstance(doc_type, pylastica.doc_type.DocType):
            doc_type = doc_type.name
        return doc_type in self._types

    @property
    def query(self):
        """

        @return:
        @rtype: pylastica.query.Query
        """
        if self._query is None:
            self._query = pylastica.query.Query.create('')
        return self._query

    @query.setter
    def query(self, query):
        """
        @param query:
        @type query: str or dict or pylastica.query.Query
        """
        self.set_query(query)

    @classmethod
    def create(cls, search_object):
        """
        Create a new search object
        @param cls:
        @type cls:
        @param search_object:
        @type search_object: pylastica.searchable.Searchable
        @return:
        @rtype: self
        """
        return search_object.create_search()

    @property
    def path(self):
        """
        Combines indices and types to get the search request path
        @return: search path (url)
        @rtype: str
        """
        indices = self.indices
        path = ''
        doc_types = self.doc_types
        if len(indices) == 0:
            if len(doc_types) != 0:
                path += '_all'
        else:
            path += ','.join(indices)
        if len(doc_types) != 0:
            path += '/' + ','.join(doc_types)
        return path + '/_search'

    def search(self, query=None, options=None):
        """
        Search in the set indices and types
        @param query:
        @type query: mixed
        @param options: optional limit or dict of options
        @type options: int or dict
        @return:
        @rtype: pylastica.resultset.ResultSet
        """
        self.set_options_and_query(options, query)
        query = self.query
        response = self.client.request(self.path, data=query.to_dict(), query=self.options)
        return pylastica.resultset.ResultSet(response, query)

    def count(self, query=None):
        """
        Counts results for a query. If no query is set, a MatchAll query is used.
        @param query: dict with all query data or a Query object
        @type query: dict or pylastica.query.Query
        @return: number of docs matching the query
        @rtype: int
        """
        self.set_options_and_query(None, query)
        query = self.query
        response = self.client.request(self.path, data=query.to_dict(), query={self.OPTION_SEARCH_TYPE: self.SEARCH_TYPE_COUNT})
        return pylastica.resultset.ResultSet(response, query).get_total_hits()

    def set_options_and_query(self, options=None, query=None):
        """

        @param options:
        @type options: int or dict
        @param query:
        @type query: str or dict or pylastica.query.Query
        @return:
        @rtype: self
        """
        if query is not None:
            self.query = query
        if isinstance(options, int):
            self._query.size = options
        elif isinstance(options, dict):
            if 'limit' in options:
                self._query.size = options['limit']
                del options['limit']
            if 'explain' in options:
                self._query.explain = options['explain']
                del options['explain']
            self.set_options(options)
        return self



__author__ = 'Joe Linn'


from .abstract import *
from .bool import *
from .common import *
from .constantscore import *
from .customscore import *
from .filtered import *
from .functionscore import *
from .fuzzy import *
from .fuzzylikethis import *
from .haschild import *
from .hasparent import *
from .ids import *
from .match import *
from .matchall import *
from .morelikethis import *
from .multimatch import *
from .nested import *
from .prefix import *
from .query import *
from .querystring import *
from .range import *
from .simple import *
from .simplequerystring import SimpleQueryString
from .term import *
from .terms import *
from .wildcard import *




__author__ = 'Joe Linn'

from .abstract import AbstractQuery


class Nested(AbstractQuery):
    SCORE_MODE_AVG = 'avg'
    SCORE_MODE_TOTAL = 'total'
    SCORE_MODE_MAX = 'max'
    SCORE_MODE_NONE = 'none'

    def set_path(self, path):
        """
        Set the path for the nested query
        @param path:
        @type path: str
        @return:
        @rtype: eslf
        """
        return self.set_param('path', path)

    def set_query(self, query):
        """
        Set the query for this nested query
        @param query:
        @type query: pylastica.query.AbstractQuery
        @return:
        @rtype: self
        """
        assert isinstance(query, AbstractQuery), "query must be an instance of an implementation of AbstractQuery: %r" % query
        return self.set_param('query', query.to_dict())

    def set_score_mode(self, mode):
        """
        Set score method
        @param mode: options: avg, total, max, none
        @type mode: str
        @return:
        @rtype: self
        """
        return self.set_param('score_mode', mode)



__author__ = 'Joe Linn'

import abc

class Searchable(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def search(self, query=None, options=None):
        """
        Searches results for a query
        @param query: dict with all query data or a Query object
        @type query: str or dict or pylastica.query.Query
        @param options:
        @type options: dict
        @return: result set with all results
        @rtype: pylastica.resultset.ResultSet
        """
        pass

    @abc.abstractmethod
    def count(self, query=None):
        """
        Counts results for a query. If no query is set, a MatchAll query is used.
        @param query: dict with all query data or a Query object
        @type query: dict or pylastica.query.Query
        @return: number of docs matching the query
        @rtype: int
        """
        pass

    @abc.abstractmethod
    def create_search(self, query=None, options=None):
        """

        @param query:
        @type query: pylastica.query.Query
        @param options:
        @type options:  dict
        @return:
        @rtype: pylastica.search.Search
        """
        pass



__author__ = 'Joe Linn'

from .indexdocument import IndexDocument


class CreateDocument(IndexDocument):
    def __init__(self, document):
        super(CreateDocument, self).__init__(document)
        self._op_type = self.OP_TYPE_CREATE



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter
import pylastica.exception

class Bool(AbstractFilter):
    def __init__(self):
        self._boost = 1.0
        self._must = []
        self._should = []
        self._must_not = []
        super(Bool, self).__init__()

    def add_should(self, args):
        """
        Add should filter
        @param args:
        @type args: dict or pylastica.filter.AbstractFilter
        @return:
        @rtype: self
        """
        return self._add_filter('should', args)

    def add_must(self, args):
        """
        Add must filter
        @param args:
        @type args: class AbstractFilter(pylastica.param.Param):
        @return:
        @rtype: self
        """
        return self._add_filter('must', args)

    def add_must_not(self, args):
        """
        Add mustNot filter
        @param args:
        @type args: class AbstractFilter(pylastica.param.Param):
        @return:
        @rtype: self
        """
        return self._add_filter('must_not', args)

    def _add_filter(self, type, args):
        """
        Add a general filter based on type
        @param type: filter type
        @type type: str
        @param args:
        @type args: dict or pylastica.filter.AbstractFilter
        @return:
        @rtype: self
        """
        if isinstance(args, pylastica.filter.abstractfilter.AbstractFilter):
            args = args.to_dict()
        assert isinstance(args, dict), "Invalid parameter. Must be a dict or instance of implementation of AbstractFilter."
        var_name = '_' + type
        self.__dict__[var_name].append(args)
        return self

    def to_dict(self):
        args = {
            'bool': {}
        }
        if len(self._must):
            args['bool']['must'] = self._must
        if len(self._should):
            args['bool']['should'] = self._should
        if len(self._must_not):
            args['bool']['must_not'] = self._must_not
        return args

    def set_boost(self, boost):
        """
        Set the boost for this filter
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        self._boost = float(boost)
        return self



__author__ = 'Joe Linn'

from .abstract import AbstractQuery

class Match(AbstractQuery):
    def set_field(self, field, values):
        """
        Set the param for the message array
        @param field:
        @type field: str
        @param values:
        @type values: mixed
        @return:
        @rtype: self
        """
        return self.set_param(field, values)

    def set_field_param(self, field, key, value):
        """
        Set a param for the given field
        @param field: field name
        @type field: str
        @param key:
        @type key: str
        @param value:
        @type value: mixed
        @return:
        @rtype: self
        """
        if field not in self._params:
            self._params[field] = {}
        self._params[field][key] = value
        return self

    def set_field_query(self, field, query):
        """
        Set the query string for a field
        @param field:
        @type field: str
        @param query:
        @type query: str
        @return:
        @rtype: self
        """
        return self.set_field_param(field, 'query', query)

    def set_field_type(self, field, doc_type):
        """
        Set the doc type for a field
        @param field:
        @type field: str
        @param doc_type:
        @type doc_type: str
        @return:
        @rtype: self
        """
        return self.set_field_param(field, 'type', doc_type)

    def set_field_operator(self, field, operator):
        """
        Set the field operator
        @param field:
        @type field: str
        @param operator:
        @type operator: str
        @return:
        @rtype: self
        """
        return self.set_field_param(field, 'operator', operator)

    def set_field_analyzer(self, field, analyzer):
        """
        Set the field analyzer
        @param field:
        @type field: str
        @param analyzer:
        @type analyzer: str
        @return:
        @rtype: self
        """
        return self.set_field_param(field, 'anaylzer', analyzer)

    def set_field_boost(self, field, boost=1.0):
        """
        Set the boost for a field
        @param field:
        @type field: str
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        return self.set_field_param(field, 'boost', float(boost))

    def set_field_minimum_should_match(self, field, minimum):
        """
        Set the filed minimum should match
        @param field:
        @type field: str
        @param minimum:
        @type minimum: int
        @return:
        @rtype: self
        """
        return self.set_field_param(field, 'minimum_should_match', int(minimum))

    def set_field_fuzziness(self, field, fuzziness):
        """
        Set field fuzziness
        @param field:
        @type field: str
        @param fuzziness:
        @type fuzziness: float
        @return:
        @rtype: self
        """
        return self.set_field_param(field, 'fuzziness', float(fuzziness))

    def set_field_fuzzy_rewrite(self, field, fuzzy_rewrite):
        """
        Set field fuzzy rewrite
        @param field:
        @type field: str
        @param fuzzy_rewrite:
        @type fuzzy_rewrite: str
        @return:
        @rtype: self
        """
        return self.set_field_param(field, 'fuzzy_rewrite', fuzzy_rewrite)

    def set_field_prefix_length(self, field, prefix_length):
        """
        Set field prefix length
        @param field:
        @type field: str
        @param prefix_length:
        @type prefix_length: int
        @return:
        @rtype: self
        """
        return self.set_field_param(field, 'prefix_length', int(prefix_length))

    def set_field_max_expansions(self, field, max_expansions):
        """
        Set field max expansions
        @param field:
        @type field: str
        @param max_expansions:
        @type max_expansions: int
        @return:
        @rtype: self
        """
        return self.set_field_param(field, 'max_expansions', int(max_expansions))



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter

class BoolNot(AbstractFilter):
    def __init__(self, filter_object):
        super(BoolNot, self).__init__()
        self.set_filter(filter_object)

    def set_filter(self, filter_object):
        """
        Set filter
        @param filter_object:
        @type filter_object: pylastica.filter.AbstractFilter
        @return:
        @rtype: self
        """
        assert isinstance(filter_object, AbstractFilter), "filter_object must be in instance of an implementation of AbstractFilter: %r" % filter_object
        return self.set_param('filter',filter_object.to_dict())

    def _get_base_name(self):
        """

        @return:
        @rtype: str
        """
        return 'not'



__author__ = 'Joe Linn'

from .abstractdocument import AbstractDocument


class DeleteDocument(AbstractDocument):
    def __init__(self, document):
        super(DeleteDocument, self).__init__(document)
        self._op_type = self.OP_TYPE_DELETE

    def _get_metadata_by_document(self, document):
        """

        @param document:
        @type document: pylastica.document.Document
        @return:
        @rtype: dict
        """
        params = [
            'index',
            'type',
            'id',
            'version',
            'version_type',
            'routing',
            'parent'
        ]
        return document.get_options(params, True)



__author__ = 'Joe Linn'

from .abstract import AbstractQuery


class MoreLikeThis(AbstractQuery):
    def set_fields(self, fields):
        """
        Set fields for this query
        @param fields:
        @type fields: list of str
        @return:
        @rtype: self
        """
        return self.set_param('fields', fields)

    def set_like_text(self, text):
        """
        Set the like_text value
        @param text:
        @type text: str
        @return:
        @rtype: self
        """
        return self.set_param('like_text', text.strip())

    def set_boost(self, boost):
        """
        Set the boost
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        return self.set_param('boost', float(boost))

    def set_max_query_terms(self, max_query_terms):
        """
        Set max_query_terms
        @param max_query_terms:
        @type max_query_terms: int
        @return:
        @rtype: self
        """
        return self.set_param('max_query_terms', int(max_query_terms))

    def set_percent_terms_to_match(self, percent):
        """
        Set percent terms to match
        @param percent:
        @type percent: float
        @return:
        @rtype: self
        """
        return self.set_param('percent_terms_to_match', float(percent))

    def set_min_term_frequency(self, min_frequency):
        """
        Set minimum term frequency
        @param min_frequency:
        @type min_frequency: int
        @return:
        @rtype: self
        """
        return self.set_param('min_term_freq', int(min_frequency))

    def set_min_doc_frequency(self, min_freq):
        """
        Set minimum document frequency
        @param min_freq:
        @type min_freq: int
        @return:
        @rtype: self
        """
        return self.set_param('min_doc_freq', int(min_freq))

    def set_max_doc_frequency(self, max_freq):
        """
        Set the maximum doc frequency
        @param max_freq:
        @type max_freq: int
        @return:
        @rtype: self
        """
        return self.set_param('max_doc_freq', int(max_freq))

    def set_min_word_length(self, length):
        """
        Set minimum word length
        @param length:
        @type length: int
        @return:
        @rtype: self
        """
        return self.set_param('min_word_length', int(length))

    def set_max_word_length(self, length):
        """
        Set maximum word length
        @param length:
        @type length: int
        @return:
        @rtype: str
        """
        return self.set_param('max_word_length', length)

    def set_boost_terms(self, terms):
        """
        Set boost terms
        @param terms:
        @type terms: bool
        @return:
        @rtype: self
        """
        return self.set_param('boost_terms', bool(terms))

    def set_analyzer(self, analyzer):
        """
        Set analyzer
        @param analyzer:
        @type analyzer: str
        @return:
        @rtype: self
        """
        return self.set_param('analyzer', analyzer.strip())

    def set_stop_words(self, stop_words):
        """
        Set stop words
        @param stop_words:
        @type stop_words: list of str
        @return:
        @rtype: self
        """
        assert isinstance(stop_words, list), "stop_words must be a list: %r" % stop_words
        return self.set_param('stop_words', stop_words)



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter

class Prefix(AbstractFilter):
    def __init__(self, field='', prefix=''):
        """
        @param field: field name
        @type field: str
        @param prefix: prefix string
        @type prefix: str
        """
        super(Prefix, self).__init__()
        self.set_field(field)
        self.set_prefix(prefix)

    def set_field(self, field):
        """
        Set the name of the prefix field
        @param field: field name
        @type field: str
        @return:
        @rtype: self
        """
        self._field = field
        return self

    def set_prefix(self, prefix):
        """
        Set the prefix string
        @param prefix:
        @type prefix: str
        @return:
        @rtype: self
        """
        self._prefix = prefix
        return self

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        self.set_param(self._field, self._prefix)
        return super(Prefix, self).to_dict()





__author__ = 'Joe Linn'

import re
import dateutil.parser
import importlib

def get_param_name(name_or_object):
    """
    Attempts to determine the name of the parameter based on its class
    @param name_or_object: class instance or class name
    @type name_or_object: str or object
    @return: parameter name
    @rtype: str
    """
    if not isinstance(name_or_object, str) and isinstance(name_or_object, object):
        name_or_object = name_or_object.__class__.__name__
    name_or_object = re.sub(r'^(Facet|Query|Filter)|(Facet|Query|Filter)$', r'', name_or_object)
    return to_snake_case(name_or_object)

def to_snake_case(string):
    """
    Converts a CamelCase string to snake_case
    @param string:
    @type string: str
    @return:
    @rtype: str
    """
    string = re.sub(r'([A-Z])', r'_\1', string)
    return string[1:].lower()

def get_class(kls):
    parts = kls.split('.')
    module = ".".join(parts[:-1]) + '.' + parts[-1].lower()
    #m = __import__( module, level=0)
    m = importlib.import_module(module)
    # for comp in parts[1:]:
    #     m = getattr(m, comp)
    m = getattr(m, parts[-1])
    return m

def convert_date(date):
    return dateutil.parser.parse(date).strftime('%s')



__author__ = 'Joe Linn'

import json
#import pylastica
import pylastica.param

class Request(pylastica.param.Param):
    POST = 'POST'
    PUT = 'PUT'
    GET = 'GET'
    DELETE = 'DELETE'

    def __init__(self, path, method='GET', data=None, query=None, connection=None):
        """
        @param path: Request path
        @type path: str
        @param method: optional request method (use class properties)
        @type method: str
        @param data: optional request data
        @type data: dict
        @param query: optional query params
        @type query: dict
        @param connection: optional connection object
        @type connection: pylastica.connection.Connection
        """
        super(Request, self).__init__()
        self._connection = None
        if not query: query = {}
        if not data: data = {}
        self.set_path(path)
        self.set_method(method)
        self.set_data(data)
        self.set_query(query)
        if connection is not None:
            self.set_connection(connection)

    def set_method(self, method):
        """
        Set the request method. Use one of the class properties.
        @param method: request method
        @type method: str
        @return:
        @rtype: self
        """
        return self.set_param('method', method)

    def get_method(self):
        """
        Get the request method
        @return:
        @rtype: str
        """
        return self.get_param('method')

    def set_data(self, data):
        """
        Set the request data
        @param data:
        @type data: dict
        @return:
        @rtype: self
        """
        return self.set_param('data', data)

    def get_data(self):
        """
        Get request data
        @return:
        @rtype: dict
        """
        return self.get_param('data')

    def set_path(self, path):
        """
        Set the request path
        @param path:
        @type path: str
        @return:
        @rtype: self
        """
        return self.set_param('path', path)

    def get_path(self):
        """
        Get the request path
        @return:
        @rtype: str
        """
        return self.get_param('path')

    @property
    def path(self):
        """

        @return:
        @rtype: str
        """
        return self.get_path()

    @path.setter
    def path(self, path):
        """

        @param path:
        @type path: str
        """
        self.set_path(path)

    def set_query(self, query):
        """

        @param query:
        @type query: dict
        @return:
        @rtype: self
        """
        return self.set_param('query', query)

    def get_query(self):
        """
        Retrun query params
        @return:
        @rtype: dict
        """
        return self.get_param('query')

    @property
    def query(self):
        """

        @return:
        @rtype: dict
        """
        return self.get_query()

    @query.setter
    def query(self, query):
        """

        @param query:
        @type query: dict
        """
        self.set_query(query)

    def set_connection(self, connection):
        """

        @param connection:
        @type connection: pylastica.connection.Connection
        @return:
        @rtype: self
        """
        assert isinstance(connection, pylastica.connection.Connection), "connection must be of type Connection: %r" % connection
        self._connection = connection
        return self

    def get_connection(self):
        """
        Return Connection object
        @return:
        @rtype: pylastica.connection.Connection
        """
        if self._connection is None:
            raise pylastica.exception.InvalidException("No valid connection object set.")
        return self._connection

    def send(self):
        """
        Send a request to the server
        @return:
        @rtype: pylastica.response.Response
        """
        transport = self.get_connection().get_transport_object()
        return transport.execute(self, self.get_connection().to_dict())

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        data = self.params
        if self._connection is not None:
            data['connection'] = self._connection.get_params()
        return data

    def to_str(self):
        """
        Convert request to curl request format
        @return:
        @rtype: str
        """
        return json.dumps(self.to_dict())

    def __str__(self):
        return self.to_str()



__author__ = 'Joe Linn'

import pylastica.response
import pylastica.result

class ResultSet(object):
    def __init__(self, response, query):
        """
        @type response: pylastica.response.Response
        @type query: pylastica.query.Query
        """
        assert isinstance(response, pylastica.response.Response), "response must be of type Response: %r" % response
        assert isinstance(query, pylastica.query.Query), "query must be of type Query: %r" % query
        self._results = []
        self._response = response
        self._query = query
        self._took = 0
        self._timed_out = False
        self._total_hits = 0
        self._max_score = 0
        self._init(response)

    def _init(self, response):
        """
        Loads all data into the results object
        @type response: pylastica.response.Response
        """
        result = response.get_data()
        if 'hits' in result:
            self._total_hits = result['hits']['total'] if 'total' in result['hits'] else 0
            self._max_score = result['hits']['max_score'] if 'max_score' in result['hits'] else 0
            self._took = result['took'] if 'took' in result else 0
            self._timed_out = 'timed_out' in result and result['timed_out'] is not None
            if 'hits' in result['hits']:
                self._results = [pylastica.result.Result(hit) for hit in result['hits']['hits']]

    def get_results(self):
        """
        Returns all results
        @rtype: list
        """
        return self._results

    @property
    def results(self):
        """

        @return:
        @rtype: list of pylastica.result.Result
        """
        return self.get_results()

    def has_facets(self):
        """
        Determine whether or not this result set has facets
        @rtype: bool
        """
        data = self._response.get_data()
        return 'facets' in data

    def get_facets(self):
        """
        Returns all facet results
        @return:
        @rtype: dict
        """
        data = self._response.get_data()
        return data['facets'] if 'facets' in data else {}

    def has_aggregations(self):
        """
        Determine whether or not this result set has aggregations
        @return:
        @rtype: bool
        """
        data = self._response.get_data()
        return 'aggregations' in data

    @property
    def aggregations(self):
        """

        @return:
        @rtype: dict
        """
        data = self._response.get_data()
        return data['aggregations'] if 'aggregations' in data else {}

    def get_total_hits(self):
        """
        Returns the number of total found hits
        @return:
        @rtype: int
        """
        return int(self._total_hits)

    def get_max_score(self):
        """
        Returns the max score of the results found
        @return:
        @rtype: float
        """
        return float(self._max_score)

    def get_total_time(self):
        """
        Returns the total time taken for this search to complete
        @return: time in milliseconds
        @rtype: int
        """
        return int(self._took)

    def has_timed_out(self):
        """
        Returns true if the query has timed out
        @return:
        @rtype: bool
        """
        return bool(self._timed_out)

    @property
    def response(self):
        """
        Returns the response object
        @return:
        @rtype: pylastica.response.Response
        """
        return self._response

    @property
    def query(self):
        """
        Return the query object
        @return:
        @rtype: pylastica.query.Query
        """
        return self._query

    @property
    def suggests(self):
        """
        Return suggestions
        @return:
        @rtype: dict
        """
        data = self._response.get_data()
        return data['suggest'] if 'suggest' in data else {}

    def has_suggests(self):
        """
        @return: true if this result set contains suggestions
        @rtype: bool
        """
        return 'suggest' in self._response.data

    def __len__(self):
        """
        Returns the size of the current result set
        @return:
        @rtype: int
        """
        return len(self._results)

    def __iter__(self):
        return iter(self._results)

    def __getitem__(self, item):
        """

        @param item:
        @type item: int
        @return:
        @rtype: pylastica.result.Result
        """
        return self._results[item]



__author__ = 'Joe Linn'

import pylastica.aggregation.abstractaggregation as abstract


class Terms(abstract.SimpleAggregation):
    """
    Terms Aggregation
    @see http://www.elasticsearch.org/guide/en/elasticsearch/reference/master/search-aggregations-bucket-terms-aggregation.html
    """
    def set_order(self, order, direction):
        """
        Set the bucket sort order
        @param order: _count, _term, or the name of a sub-aggregation or sub-aggregation response field
        @type order: str
        @param direction: asc or desc
        @type direction: str
        @return:
        @rtype: Terms
        """
        return self.set_param('order', {order: direction})

    def set_minimum_document_count(self, count):
        """
        Set the minimum number of documents in which a term must appear in order to be returned in a bucket
        @param count:
        @type count: int
        @return:
        @rtype: Terms
        """
        return self.set_param('min_doc_count', count)

    def set_include(self, pattern, flags=None):
        """
        Filter documents to include based on a regular expression
        @param pattern:
        @type pattern: str
        @param flags: Java Pattern flags
        @type flags: str
        @return:
        @rtype: Terms
        """
        if flags is None:
            return self.set_param("include", pattern)
        return self.set_param("include", {
            "pattern": pattern,
            "flags": flags
        })

    def set_exclude(self, pattern, flags=None):
        """
        Filter documents to exclude based on a regular expression
        @param pattern:
        @type pattern: str
        @param flags: Java Pattern flags
        @type flags: str
        @return:
        @rtype: Terms
        """
        if flags is None:
            return self.set_param("exclude", pattern)
        return self.set_param("exclude", {
            "pattern": pattern,
            "flags": flags
        })

    def set_execution_hint(self, hint):
        """
        Direct Elasticsearch to use direct field data or ordinals of the field values to execute this aggregation.
        The execution hint will be ignored if it is not applicable.
        @param hint: map or ordinals
        @type hint: str
        @return:
        @rtype: Terms
        """
        return self.set_param("execution_hint", hint)


#
# Autogenerated by Thrift
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#

from thrift.Thrift import *

from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
    from thrift.protocol import fastbinary
except ImportError:
    fastbinary = None


class Method(object):
    GET = 0
    PUT = 1
    POST = 2
    DELETE = 3
    HEAD = 4
    OPTIONS = 5

    _VALUES_TO_NAMES = {
        0: "GET",
        1: "PUT",
        2: "POST",
        3: "DELETE",
        4: "HEAD",
        5: "OPTIONS",
        }

    _NAMES_TO_VALUES = {
        "GET": 0,
        "PUT": 1,
        "POST": 2,
        "DELETE": 3,
        "HEAD": 4,
        "OPTIONS": 5,
        }

class Status(object):
    CONT = 100
    SWITCHING_PROTOCOLS = 101
    OK = 200
    CREATED = 201
    ACCEPTED = 202
    NON_AUTHORITATIVE_INFORMATION = 203
    NO_CONTENT = 204
    RESET_CONTENT = 205
    PARTIAL_CONTENT = 206
    MULTI_STATUS = 207
    MULTIPLE_CHOICES = 300
    MOVED_PERMANENTLY = 301
    FOUND = 302
    SEE_OTHER = 303
    NOT_MODIFIED = 304
    USE_PROXY = 305
    TEMPORARY_REDIRECT = 307
    BAD_REQUEST = 400
    UNAUTHORIZED = 401
    PAYMENT_REQUIRED = 402
    FORBIDDEN = 403
    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    NOT_ACCEPTABLE = 406
    PROXY_AUTHENTICATION = 407
    REQUEST_TIMEOUT = 408
    CONFLICT = 409
    GONE = 410
    LENGTH_REQUIRED = 411
    PRECONDITION_FAILED = 412
    REQUEST_ENTITY_TOO_LARGE = 413
    REQUEST_URI_TOO_LONG = 414
    UNSUPPORTED_MEDIA_TYPE = 415
    REQUESTED_RANGE_NOT_SATISFIED = 416
    EXPECTATION_FAILED = 417
    UNPROCESSABLE_ENTITY = 422
    LOCKED = 423
    FAILED_DEPENDENCY = 424
    INTERNAL_SERVER_ERROR = 500
    NOT_IMPLEMENTED = 501
    BAD_GATEWAY = 502
    SERVICE_UNAVAILABLE = 503
    GATEWAY_TIMEOUT = 504
    INSUFFICIENT_STORAGE = 506

    _VALUES_TO_NAMES = {
        100: "CONT",
        101: "SWITCHING_PROTOCOLS",
        200: "OK",
        201: "CREATED",
        202: "ACCEPTED",
        203: "NON_AUTHORITATIVE_INFORMATION",
        204: "NO_CONTENT",
        205: "RESET_CONTENT",
        206: "PARTIAL_CONTENT",
        207: "MULTI_STATUS",
        300: "MULTIPLE_CHOICES",
        301: "MOVED_PERMANENTLY",
        302: "FOUND",
        303: "SEE_OTHER",
        304: "NOT_MODIFIED",
        305: "USE_PROXY",
        307: "TEMPORARY_REDIRECT",
        400: "BAD_REQUEST",
        401: "UNAUTHORIZED",
        402: "PAYMENT_REQUIRED",
        403: "FORBIDDEN",
        404: "NOT_FOUND",
        405: "METHOD_NOT_ALLOWED",
        406: "NOT_ACCEPTABLE",
        407: "PROXY_AUTHENTICATION",
        408: "REQUEST_TIMEOUT",
        409: "CONFLICT",
        410: "GONE",
        411: "LENGTH_REQUIRED",
        412: "PRECONDITION_FAILED",
        413: "REQUEST_ENTITY_TOO_LARGE",
        414: "REQUEST_URI_TOO_LONG",
        415: "UNSUPPORTED_MEDIA_TYPE",
        416: "REQUESTED_RANGE_NOT_SATISFIED",
        417: "EXPECTATION_FAILED",
        422: "UNPROCESSABLE_ENTITY",
        423: "LOCKED",
        424: "FAILED_DEPENDENCY",
        500: "INTERNAL_SERVER_ERROR",
        501: "NOT_IMPLEMENTED",
        502: "BAD_GATEWAY",
        503: "SERVICE_UNAVAILABLE",
        504: "GATEWAY_TIMEOUT",
        506: "INSUFFICIENT_STORAGE",
        }

    _NAMES_TO_VALUES = {
        "CONT": 100,
        "SWITCHING_PROTOCOLS": 101,
        "OK": 200,
        "CREATED": 201,
        "ACCEPTED": 202,
        "NON_AUTHORITATIVE_INFORMATION": 203,
        "NO_CONTENT": 204,
        "RESET_CONTENT": 205,
        "PARTIAL_CONTENT": 206,
        "MULTI_STATUS": 207,
        "MULTIPLE_CHOICES": 300,
        "MOVED_PERMANENTLY": 301,
        "FOUND": 302,
        "SEE_OTHER": 303,
        "NOT_MODIFIED": 304,
        "USE_PROXY": 305,
        "TEMPORARY_REDIRECT": 307,
        "BAD_REQUEST": 400,
        "UNAUTHORIZED": 401,
        "PAYMENT_REQUIRED": 402,
        "FORBIDDEN": 403,
        "NOT_FOUND": 404,
        "METHOD_NOT_ALLOWED": 405,
        "NOT_ACCEPTABLE": 406,
        "PROXY_AUTHENTICATION": 407,
        "REQUEST_TIMEOUT": 408,
        "CONFLICT": 409,
        "GONE": 410,
        "LENGTH_REQUIRED": 411,
        "PRECONDITION_FAILED": 412,
        "REQUEST_ENTITY_TOO_LARGE": 413,
        "REQUEST_URI_TOO_LONG": 414,
        "UNSUPPORTED_MEDIA_TYPE": 415,
        "REQUESTED_RANGE_NOT_SATISFIED": 416,
        "EXPECTATION_FAILED": 417,
        "UNPROCESSABLE_ENTITY": 422,
        "LOCKED": 423,
        "FAILED_DEPENDENCY": 424,
        "INTERNAL_SERVER_ERROR": 500,
        "NOT_IMPLEMENTED": 501,
        "BAD_GATEWAY": 502,
        "SERVICE_UNAVAILABLE": 503,
        "GATEWAY_TIMEOUT": 504,
        "INSUFFICIENT_STORAGE": 506,
        }


class RestRequest(object):
    """
    Attributes:
     - method
     - uri
     - parameters
     - headers
     - body
    """

    thrift_spec = (
        None, # 0
        (1, TType.I32, 'method', None, None, ), # 1
        (2, TType.STRING, 'uri', None, None, ), # 2
        (3, TType.MAP, 'parameters', (TType.STRING,None,TType.STRING,None), None, ), # 3
        (4, TType.MAP, 'headers', (TType.STRING,None,TType.STRING,None), None, ), # 4
        (5, TType.STRING, 'body', None, None, ), # 5
    )

    def __init__(self, method=None, uri=None, parameters=None, headers=None, body=None):
        self.method = method
        self.uri = uri
        self.parameters = parameters
        self.headers = headers
        self.body = body

    def read(self, iprot):
        if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
            fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.method = iprot.readI32();
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.uri = iprot.readString();
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.MAP:
                    self.parameters = {}
                    (_ktype1, _vtype2, _size0 ) = iprot.readMapBegin()
                    for _i4 in xrange(_size0):
                        _key5 = iprot.readString();
                        _val6 = iprot.readString();
                        self.parameters[_key5] = _val6
                    iprot.readMapEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.MAP:
                    self.headers = {}
                    (_ktype8, _vtype9, _size7 ) = iprot.readMapBegin()
                    for _i11 in xrange(_size7):
                        _key12 = iprot.readString();
                        _val13 = iprot.readString();
                        self.headers[_key12] = _val13
                    iprot.readMapEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.STRING:
                    self.body = iprot.readString();
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
            oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('RestRequest')
        if self.method != None:
            oprot.writeFieldBegin('method', TType.I32, 1)
            oprot.writeI32(self.method)
            oprot.writeFieldEnd()
        if self.uri != None:
            oprot.writeFieldBegin('uri', TType.STRING, 2)
            oprot.writeString(self.uri)
            oprot.writeFieldEnd()
        if self.parameters != None:
            oprot.writeFieldBegin('parameters', TType.MAP, 3)
            oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.parameters))
            for kiter14,viter15 in self.parameters.items():
                oprot.writeString(kiter14)
                oprot.writeString(viter15)
            oprot.writeMapEnd()
            oprot.writeFieldEnd()
        if self.headers != None:
            oprot.writeFieldBegin('headers', TType.MAP, 4)
            oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.headers))
            for kiter16,viter17 in self.headers.items():
                oprot.writeString(kiter16)
                oprot.writeString(viter17)
            oprot.writeMapEnd()
            oprot.writeFieldEnd()
        if self.body != None:
            oprot.writeFieldBegin('body', TType.STRING, 5)
            oprot.writeString(self.body)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()
        def validate(self):
            if self.method is None:
                raise TProtocol.TProtocolException(message='Required field method is unset!')
            if self.uri is None:
                raise TProtocol.TProtocolException(message='Required field uri is unset!')
            return


    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.iteritems()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)

class RestResponse(object):
    """
    Attributes:
     - status
     - headers
     - body
    """

    thrift_spec = (
        None, # 0
        (1, TType.I32, 'status', None, None, ), # 1
        (2, TType.MAP, 'headers', (TType.STRING,None,TType.STRING,None), None, ), # 2
        (3, TType.STRING, 'body', None, None, ), # 3
    )

    def __init__(self, status=None, headers=None, body=None):
        self.status = status
        self.headers = headers
        self.body = body

    def read(self, iprot):
        if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
            fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.status = iprot.readI32();
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.MAP:
                    self.headers = {}
                    (_ktype19, _vtype20, _size18 ) = iprot.readMapBegin()
                    for _i22 in xrange(_size18):
                        _key23 = iprot.readString();
                        _val24 = iprot.readString();
                        self.headers[_key23] = _val24
                    iprot.readMapEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.body = iprot.readString();
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
            oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
            return
        oprot.writeStructBegin('RestResponse')
        if self.status != None:
            oprot.writeFieldBegin('status', TType.I32, 1)
            oprot.writeI32(self.status)
            oprot.writeFieldEnd()
        if self.headers != None:
            oprot.writeFieldBegin('headers', TType.MAP, 2)
            oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.headers))
            for kiter25,viter26 in self.headers.items():
                oprot.writeString(kiter25)
                oprot.writeString(viter26)
            oprot.writeMapEnd()
            oprot.writeFieldEnd()
        if self.body != None:
            oprot.writeFieldBegin('body', TType.STRING, 3)
            oprot.writeString(self.body)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()
        def validate(self):
            if self.status is None:
                raise TProtocol.TProtocolException(message='Required field status is unset!')
            return


    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.iteritems()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


__author__ = 'Joe Linn'

import abc
from .action import Action
from pylastica.document import Document
from pylastica.script import Script


class AbstractDocument(Action):
    __metaclass__ = abc.ABCMeta

    def __init__(self, document):
        """

        @param document:
        @type document: pylastica.document.Document or pylastica.script.Script
        """
        super(AbstractDocument, self).__init__()
        self.data = document

    @property
    def document(self):
        """

        @return:
        @rtype: pylastica.document.Document
        """
        return self.get_document()

    @document.setter
    def document(self, document):
        """

        @param document:
        @type document: pylastica.document.Document
        """
        self.set_document(document)

    @property
    def script(self):
        """

        @return:
        @rtype: pylastica.script.Script
        """
        if not isinstance(self._data, Script):
            raise TypeError("This action does not contain a Script: %r" % self._data)
        return self._data

    @script.setter
    def script(self, script):
        """

        @param script:
        @type script: pylastica.script.Script
        """
        self.set_script(script)

    def set_script(self, script):
        """

        @param script:
        @type script: pylastica.script.Script
        @return:
        @rtype: self
        """
        if not isinstance(script, Script):
            raise TypeError("script must be an instance of Script: %r" % script)
        # elif self.op_type != self.OP_TYPE_UPDATE:
        #     raise NotImplementedError("scripts may only be set when performing an update operation: %r" % self.op_type)
        self._data = script
        metadata = self._get_metadata_by_script(script)
        self.metadata = metadata
        return self

    def set_document(self, document):
        """

        @param document:
        @type document: pylastica.document.Document
        @return:
        @rtype: self
        """
        import pylastica.document
        assert isinstance(document, pylastica.document.Document), "document must be an instance of Document: %r" % document
        self._data = document
        metadata = self._get_metadata_by_document(document)
        self.metadata = metadata
        return self

    def get_document(self):
        """

        @return:
        @rtype: pylastica.document.Document
        """
        if not isinstance(self._data, Document):
            raise TypeError("This action does not contain a Document: %r" % self._data)
        return self._data

    def set_data(self, data):
        """

        @param data:
        @type data: pylastica.document.Document or pylastica.script.Script
        @return:
        @rtype: self
        """
        if isinstance(data, Script):
            self.script = data
        elif isinstance(data, Document):
            self.document = data
        else:
            raise TypeError("data must be of type Document or Script: %r" % data)
        return self

    def get_data(self):
        """

        @return:
        @rtype: pylastica.document.Document or pylastica.script.Script
        """
        return self._data

    @property
    def data(self):
        """

        @return:
        @rtype: pylastica.document.Document or pylastica.script.Script
        """
        return self.get_data()

    @data.setter
    def data(self, data):
        """

        @param data:
        @type data: pylastica.document.Document or pylastica.script.Script
        """
        self.set_data(data)

    @abc.abstractmethod
    def _get_metadata_by_document(self, document):
        """

        @param document:
        @type document: pylastica.document.Document
        @return:
        @rtype: dict
        """
        pass

    def _get_metadata_by_script(self, script):
        """

        @param script:
        @type script: pylastica.script.Script
        @return:
        @rtype: dict
        """
        pass

    @classmethod
    def create(cls, data, op_type=None):
        """

        @param cls:
        @type cls:
        @param data:
        @type data: pylastica.document.Document or pylastica.script.Script
        @param op_type: bulk operation type
        @type op_type: str
        @return:
        @rtype: cls
        """
        if not isinstance(data, Document) and not isinstance(data, Script):
            raise TypeError("data must be an instance of Document or Script: %r" % data)
        if op_type is None and data.has_op_type():
            op_type = data.op_type
        if isinstance(data, Script) and op_type is not None and op_type != cls.OP_TYPE_UPDATE:
            raise TypeError("Scripts may only be used for updates.")

        if op_type == cls.OP_TYPE_DELETE:
            from .deletedocument import DeleteDocument
            action = DeleteDocument(data)
        elif op_type == cls.OP_TYPE_CREATE:
            from .createdocument import CreateDocument
            action = CreateDocument(data)
        elif op_type == cls.OP_TYPE_UPDATE:
            from .updatedocument import UpdateDocument
            action = UpdateDocument(data)
        else:
            from .indexdocument import IndexDocument
            action = IndexDocument(data)
        return action



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractGeoDistance

class GeoDistance(AbstractGeoDistance):
    DISTANCE_TYPE_ARC = 'arc'
    DISTANCE_TYPE_PLANE = 'plane'

    OPTIMIZE_BBOX_MEMORY = 'memory'
    OPTIMIZE_BBOX_INDEXED = 'indexed'
    OPTIMIZED_BBOX_NONE = 'none'

    def __init__(self, key, location, distance):
        """
        @param key:
        @type key: str
        @param location: location as dict or geohash {'lat':40.3, 'lon':45.2}
        @type location: dict or str
        @param distance:
        @type distance: str
        """
        super(GeoDistance, self).__init__(key, location)
        self.set_distance(distance)


    def set_distance(self, distance):
        """
        @param distance:
        @type distance: str
        @return:
        @rtype: self
        """
        return self.set_param('distance', distance)

    def set_distance_type(self, distance_type):
        """
        See DISTANCE_TYPE_* class properties
        @param distance_type:
        @type distance_type: str
        @return:
        @rtype: self
        """
        return self.set_param('distance_type', distance_type)

    def set_optimize_bbox(self, optimize):
        """
        See OPTIMIZE_BBOX_* constants
        @param optimize:
        @type optimize: str
        @return:
        @rtype: self
        """
        return self.set_param('optimize_box', optimize)



__author__ = 'Joe Linn'

import json
import pylastica.exception


class Response(object):
    def __init__(self, response_string, status=None):
        """
        @param response_string: response string (json)
        @type response_string: str or dict
        @param status: http status code
        @type status: int
        """
        self._query_time = None
        self._response_string = ''
        self._error = False
        self._transfer_info = {}
        self._response = None
        self._status = status
        if isinstance(response_string, dict):
            self._response = response_string
        else:
            self._response_string = response_string

    @property
    def status(self):
        """
        Get the HTTP status code for this response
        @return:
        @rtype: str
        """
        if self._status is not None:
            return self._status
        status = ''
        response = self.data
        if 'status' in response:
            status = response['status']
        return str(status)

    def get_error(self):
        """
        Error message
        @return:
        @rtype: str
        """
        message = ''
        response = self.get_data()
        if 'error' in response:
            message = response['error']
        return message

    @property
    def error(self):
        """

        @return:
        @rtype: str
        """
        return self.get_error()

    def has_error(self):
        """
        Determine if the response has an error
        @return: True if error, False otherwise
        @rtype: bool
        """
        response = self.get_data()
        if 'error' in response:
            return True
        return False

    def is_ok(self):
        """
        Check if the query returned ok
        @return:
        @rtype: bool
        """
        data = self.get_data()
        #bulk insert checks
        if 'items' in data:
            for item in data['items']:
                if not item['index']['ok']:
                    return False
            return True
        if 'status' in data:
            if 200 <= data['status'] <= 300:
                return True
            return False
        if 200 <= self.status <= 300:
            return True
        return ('ok' in data and data['ok']) or ('acknowledged' in data and data['acknowledged'])

    def get_data(self):
        """
        Response data dict
        @return:
        @rtype: dict
        """
        if self._response is None:
            response = self._response_string
            if response == False:
                self._error = True
            else:
                temp_response = json.loads(response)
                #TODO: see what happens here if an error is returned
                if temp_response is not None:
                    response = temp_response
            if response is None or response == '':
                response = {}
            if isinstance(response, str):
                response = {'message': response}
            self._response = response
        return self._response

    @property
    def data(self):
        """
        Response data dict
        @return:
        @rtype: dict
        """
        return self.get_data()

    def get_transfer_info(self):
        """
        Get the transfer information if in DEBUG mode
        @return:
        @rtype: dict
        """
        return self._transfer_info

    def set_transfer_info(self, transfer_info):
        """
        Set the transfer info of the curl request. This is only used in DEBUG mode.
        @param transfer_info:
        @type transfer_info: dict
        @return:
        @rtype: self
        """
        self._transfer_info = transfer_info
        return self

    def get_query_time(self):
        """
        Only available if DEBUG mode is enabled
        @return:
        @rtype: float
        """
        return self._query_time

    def set_query_time(self, query_time):
        """
        Set query time
        @param query_time:
        @type query_time: float
        @return:
        @rtype: self
        """
        self._query_time = query_time
        return self

    def get_engine_time(self):
        """
        Time taken by the request
        @return:
        @rtype: int
        """
        data = self.get_data()
        if 'took' not in data:
            raise pylastica.exception.NotFoundException("unable to find the field 'took' in the response.")
        return data['took']

    def get_shard_statistics(self):
        """
        Get the _shard statistics for the response
        @return:
        @rtype: dict
        """
        data = self.get_data()
        if '_shards' not in data:
            raise pylastica.exception.NotFoundException("Unable to find the field '_shards' in the response.")
        return data['_shards']

    @property
    def scroll_id(self):
        """
        Get the _scroll value for the response
        @return:
        @rtype: str
        """
        data = self.data
        if '_scroll_id' not in data:
            raise pylastica.exception.NotFoundException("unable to find the field '_scroll_id' in the response.")
        return data['_scroll_id']



__author__ = 'Joe Linn'

from .histogram import Histogram

class DateHistogram(Histogram):
    """
    @see: http://www.elasticsearch.org/guide/reference/api/search/facets/date-histogram-facet/
    """

    def set_timezone(self, offset):
        """
        Set the time_zone parameter
        @param offset:
        @type offset: str
        @return:
        @rtype: self
        """
        return self.set_param('time_zone', offset)

    def set_factor(self, factor):
        """
        Set the conversion factor for time values
        @param factor: 1000 to convert stored seconds data to milliseconds, for example
        @type factor: int
        @return:
        @rtype: self
        """
        return self.set_param('factor', factor)

    def set_pre_offset(self, offset):
        """
        Set pre rounding offset
        @param offset: date / time offset (1h, 1d, 1s, etc.)
        @type offset: str
        @return:
        @rtype: self
        """
        return self.set_param('pre_offset', offset)

    def set_post_offset(self, offset):
        """
        Set post rounding offset
        @param offset: date / time offset (1h, 1d, 1s, etc.)
        @type offset: str
        @return:
        @rtype: self
        """
        return self.set_param('post_offset', offset)

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        self._set_facet_param('date_histogram', self._params)
        return self._facet





__author__ = 'Joe Linn'

from .abstractfacet import AbstractFacet
import pylastica.exception


class Range(AbstractFacet):
    def set_field(self, field):
        """
        Set the field for the facet
        @param field: field name
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param('field', field)

    def set_key_value_fields(self, key_field, value_field):
        """
        Set the fields by their separate key and value fields
        @param key_field:
        @type key_field: str
        @param value_field:
        @type value_field: str
        @return:
        @rtype: self
        """
        return self.set_param('key_field', key_field).set_param('value_field', value_field)

    def set_key_value_scripts(self, key_script, value_script):
        """
        SEt the key and value scripts for the facet
        @param key_script:
        @type key_script: str
        @param value_script:
        @type value_script: str
        @return:
        @rtype: self
        """
        return self.set_param('key_script', key_script).set_param('value_script', value_script)

    def set_ranges(self, ranges):
        """
        Set all ranges for this facet
        @param ranges:
        @type ranges: list of dict
        @return:
        @rtype: self
        """
        return self.set_param('ranges', ranges)

    def add_range(self, range_from=None, range_to=None):
        """
        Add a range to the facet
        @param range_from: optional. If omitted, is assumed to be boundless.
        @type range_from: int or float
        @param range_to: optional. Assumed to be boundless if omitted.
        @type range_to: int or float
        @return:
        @rtype: self
        """
        if range_from is None and range_to is None:
            raise pylastica.exception.InvalidException("At least one of range_from and range_to must be set.")
        range_dict = {}
        if range_from is not None:
            range_dict['from'] = range_from
        if range_to is not None:
            range_dict['to'] = range_to
        return self.add_param('ranges', range_dict)

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        field_types_set = 0
        if 'field' in self._params:
            field_types_set += 1
        if 'key_field' in self._params:
            field_types_set += 1
        if 'key_script' in self._params:
            field_types_set += 1
        if field_types_set == 0:
            raise pylastica.exception.InvalidException("One of field, key_field or key_script must be set.")
        elif field_types_set > 1:
            raise pylastica.exception.InvalidException("Either field, key_field, or key_script should be set, but not multiple.")
        self._set_facet_param('range', self._params)
        return super(Range, self).to_dict()






__author__ = 'Joe Linn'

from .abstract import AbstractQuery

class Wildcard(AbstractQuery):
    def __init__(self, key='', value=None, boost=1.0):
        """

        @param key: wildcard key
        @type key: str
        @param value: wildcard value
        @type value: str
        @param boost:
        @type boost: float
        """
        super(Wildcard, self).__init__()
        if key != '':
            self.set_value(key, value, boost)

    def set_value(self, key, value, boost=1.0):
        """
        Set the query expression for a key with a boost value
        @param key:
        @type key: str
        @param value:
        @type value: str
        @param boost:
        @type boost: float
        @return:
        @rtype: self
        """
        return self.set_param(key, {'value': value, 'boost': boost})



__author__ = 'Joe Linn'

from .abstractfilter import AbstractGeoDistance
import pylastica.exception


class GeoDistanceRange(AbstractGeoDistance):
    RANGE_FROM = 'from'
    RANGE_TO = 'to'
    RANGE_LT = 'lt'
    RANGE_LTE = 'lte'
    RANGE_GT = 'gt'
    RANGE_GTE = 'gte'

    RANGE_INCLUDE_LOWER = 'include_lower'
    RANGE_INCLUDE_UPPER = 'include_upper'

    def __init__(self, key, location, lower=None, upper=None, include_lower=True, include_upper=True):
        """
        @param key:
        @type key: str
        @param location: location as dict or geohash {'lat':40.3, 'lon':45.2}
        @type location: dict or str
        @param lower: lower bound of range
        @type lower: float or int or str
        @param upper: upper bound of range
        @type upper: float or int or str
        @param include_lower:
        @type include_lower: bool
        @param include_upper:
        @type include_upper: bool
        """
        super(GeoDistanceRange, self).__init__(key, location)
        self._ranges = {}
        if lower is not None or upper is not None:
            self.set_range(lower, upper, include_lower, include_upper)


    def set_range(self, lower=None, upper=None, include_lower=False, include_upper=False):
        """
        @param lower: lower bound of range
        @type lower: float or int or str
        @param upper: upper bound of range
        @type upper: float or int or str
        @param include_lower:
        @type include_lower: bool
        @param include_upper:
        @type include_upper: bool
        @return:
        @rtype: self
        """
        if lower is None and upper is None:
            raise pylastica.exception.InvalidException("Either lower or upper must be given a value.")
        value = {}
        if lower is not None:
            value[self.RANGE_FROM] = lower
            value[self.RANGE_INCLUDE_LOWER] = bool(include_lower)
        if upper is not None:
            value[self.RANGE_TO] = upper
            value[self.RANGE_INCLUDE_UPPER] = bool(include_upper)
        self._ranges = value
        return self

    def to_dict(self):
        for key, value in self._ranges.iteritems():
            self.set_param(key, value)
        return super(GeoDistanceRange, self).to_dict()





__author__ = 'Joe Linn'

from .abstract import AbstractQuery
import pylastica.exception

class Fuzzy(AbstractQuery):
    def __init__(self, field_name=None, value=None):
        """
        Set either both parameters or neither.
        @param field_name: field name
        @type field_name: str
        @param value: search string
        @type value: str
        """
        super(Fuzzy, self).__init__()
        if field_name and value:
            self.set_field(field_name, value)

    def set_field(self, field_name, value):
        """

        @param field_name:
        @type field_name: str
        @param value:
        @type value: str
        @return:
        @rtype: self
        """
        if not isinstance(value, str) or not isinstance(field_name, str):
            raise pylastica.exception.InvalidException("field_name and value parameters must be str.")
        return self.set_param(field_name, {'value': value})

    def set_field_option(self, param, value):
        """
        Set optional parameters
        @param param: option name
        @type param: str
        @param value:
        @type value:
        @return:
        @rtype:
        """
        params = self.params
        if len(params) < 1:
            raise pylastica.exception.InvalidException("No field has been set.")
        params[params.keys()[0]][param] = value
        return self.set_param(params.keys()[0], params[params.keys()[0]])



__author__ = 'Joe Linn'

import abc
import pylastica.param


class AbstractFacet(pylastica.param.Param):
    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        """
        @param name: the name of the facet
        @type name: str
        """
        super(AbstractFacet, self).__init__()
        self._facet = {}
        self.name = name

    @property
    def name(self):
        """
        Get the name of this facet
        @return:
        @rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Set the name of this facet.
        @param name:
        @type name: str
        """
        assert isinstance(name, str) and name != '', "name must be a string: %r" % name
        self._name = name

    def set_filter(self, facet_filter):
        """
        Set a filter for this facet
        @param facet_filter: a filter to apply on the facet
        @type facet_filter: pylastica.filter.AbstractFilter
        @return:
        @rtype: self
        """
        if not isinstance(facet_filter, pylastica.filter.AbstractFilter):
            raise TypeError("facet_filter must be an instance of an implementation of AbstractFilter: %r" % facet_filter)
        return self._set_facet_param('facet_filter', facet_filter.to_dict())

    def set_global(self, glob=True):
        """
        Sets the flag to either run the facet globally or bound to the current search query
        @param glob:
        @type glob: bool
        @return:
        @rtype: self
        """
        return self._set_facet_param('global', bool(glob))

    def set_nested(self, nested_path):
        """
        Set the path for nexted documents
        @param nested_path: document path
        @type nested_path: str
        @return:
        @rtype: self
        """
        return self._set_facet_param('nested', nested_path)

    def set_scope(self, scope):
        """
        Set the scope
        @param scope:
        @type scope: str
        @return:
        @rtype: self
        """
        return self._set_facet_param('scope', scope)

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        return self._facet

    def _set_facet_param(self, key, value):
        """
        Sets a param for the facet. Each facet implementation must handle its own parameters.
        @param key:
        @type key: str
        @param value:
        @type value: mixed
        @return:
        @rtype: self
        """
        self._facet[key] = value
        return self



__author__ = 'Joe Linn'

import json
import pylastica.exception
import pylastica.util


class Param(object):
    def __init__(self):
        self._params = {}
        self._rawParams = {}

    def to_dict(self):
        """
        Convert the params to a dict
        @return:
        @rtype: dict
        """
        data = {self._get_base_name(): self.params}
        if len(self._rawParams):
            data.update(self._rawParams)
        return data

    def _get_base_name(self):
        """
        Param's name
        @return:
        @rtype: str
        """
        return pylastica.util.get_param_name(self)

    def _set_raw_param(self, key, value):
        """
        Sets params not inside params dict
        @param key:
        @type key: str
        @param value:
        @type value: mixed
        @return:
        @rtype: self
        """
        self._rawParams[key] = value
        return self

    def set_param(self, key, value):
        """
        Set (overwrite) the value at the given key
        @param key: key to set
        @type key: str
        @param value: value
        @type value: mixed
        @return:
        @rtype: self
        """
        self._params[key] = value
        return self

    def add_param(self, key, value):
        """
        Add a param to the list at key
        @param key:
        @type key: str
        @param value:
        @type value: mixed
        @return:
        @rtype: self
        """
        if not key in self._params:
            self._params[key] = []
        self._params[key].append(value)
        return self

    def get_param(self, key):
        """
        Get a specific parameter
        @param key:
        @type key: str
        @return:
        @rtype: mixed
        """
        if not key in self._params:
            raise pylastica.exception.InvalidException("Param %s does not exist." % key)
        return self._params[key]

    def has_param(self, key):
        """
        Test if a param exists
        @param key:
        @type key: str
        @return:
        @rtype: bool
        """
        return key in self._params

    @property
    def params(self):
        """
        Get the params dict
        @return:
        @rtype: dict
        """
        return self._params

    @params.setter
    def params(self, params):
        """
        Set (overwrite) all params of this object
        @param params:
        @type params: dict
        """
        assert isinstance(params, dict), "params must be of type dict: %r" % params
        self._params = params

    def __str__(self):
        """
        @return:
        @rtype: str
        """
        return json.dumps(self.to_dict())

    def __unicode__(self):
        """

        @return:
        @rtype: unicode
        """
        return unicode(self.__str__())



__author__ = 'Joe Linn'

from .abstract import AbstractQuery


class Range(AbstractQuery):
    """
    @see: http://www.elasticsearch.org/guide/reference/query-dsl/range-query/
    """
    def __init__(self, field_name=None, args=None):
        """

        @param field_name:
        @type field_name: str
        @param args:
        @type args: dict
        """
        super(Range, self).__init__()
        if field_name:
            self.add_field(field_name, args)

    def add_field(self, field_name, args=None):
        """
        Add a range field to the query
        @param field_name:
        @type field_name: str
        @param args:
        @type args: dict
        @return:
        @rtype: self
        """
        if args is None:
            args = {}
        return self.set_param(field_name, args)



__author__ = 'Joe Linn'

from .abstractfacet import AbstractFacet
import pylastica.script

class Terms(AbstractFacet):
    ORDER_COUNT = 'count'
    ORDER_TERM = 'term'
    ORDER_REVERSE_COUNT = 'reverse_count'
    ORDER_REVERSE_TERM = 'reverse_term'

    def set_field(self, field):
        """
        Set the field for this facet
        @param field:
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param('field', field)

    def set_script(self, script):
        """
        Set the script for this facet
        @param script:
        @type script: pylastica.script.Script
        @return:
        @rtype: self
        """
        if not isinstance(script, pylastica.script.Script):
            raise TypeError("script must be of type Script: %r" % script)
        return self.set_param('script', script.script).set_param('params', script.params)

    def set_script_field(self, script):
        """
        Set the script field, to be used if no field is provided
        @param script:
        @type script: pylastica.script.Script
        @return:
        @rtype: self
        """
        if not isinstance(script, pylastica.script.Script):
            raise TypeError("script must be of type Script: %r" % script)
        return self.set_param('script_field', script.script).set_param('params', script.params)

    def set_fields(self, fields):
        """
        Set multiple fields for terms
        @param fields:
        @type fields: list of str
        @return:
        @rtype: self
        """
        return self.set_param('fields', fields)

    def set_all_terms(self, all_terms=True):
        """
        Sets the flag to return all available terms. If the term does not have a hit, the count is 0.
        @param all_terms:
        @type all_terms: bool
        @return:
        @rtype: self
        """
        return self.set_param('all_terms', bool(all_terms))

    def set_order(self, order):
        """
        Set the ordering for this facet
        @param order: See ORDER_* class properties for options
        @type order: str
        @return:
        @rtype: self
        """
        return self.set_param('order', order)

    def set_exclude(self, exclude):
        """
        Sets a list of terms which are excluded from the search
        @param exclude:
        @type exclude: list of str
        @return:
        @rtype: self
        """
        return self.set_param('exclude', exclude)


    def set_size(self, size):
        """
        Set the number of terms to return
        @param size:
        @type size: int
        @return:
        @rtype: self
        """
        return self.set_param('size', int(size))

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        self._set_facet_param('terms', self._params)
        return super(Terms, self).to_dict()




__author__ = 'Joe Linn'

import pylastica

class Stats(object):

    def __init__(self, index):
        """
        @param index:
        @type index: pylastica.index.Index
        """
        assert isinstance(index, pylastica.index.Index), "index must be of type Index: %r" % index
        self._index = index
        self._response = None
        self._data = {}
        self.refresh()

    @property
    def data(self):
        """
        Returns the raw stats info
        @return:
        @rtype: dict
        """
        return self._data

    def get(self, *args):
        """
        Returns data based on keys in args
        @param args: keys of data to return
        @type args: str
        @return:
        @rtype: dict or None
        """
        #TODO: this behavior doesn't seem right. If multiple args can be passed, why only return data for the first one?
        data = self.data
        for arg in args:
            if arg in data:
                data = data[arg]
            else:
                return None
        return data

    @property
    def index(self):
        """
        Return the index object
        @return:
        @rtype: pylastica.index.Index
        """
        return self._index

    @property
    def response(self):
        """
        Return the response object
        @return:
        @rtype: pylastica.response.Response
        """
        return self._response

    def refresh(self):
        """
        Reloads the status data for this object
        """
        self._response = self.index.request('_stats', pylastica.request.Request.GET)
        self._data = self.response.data



__author__ = 'Joe Linn'

class Shard(object):
    def __init__(self, shard_number, data):
        """
        @param shard_number: the shard index / number
        @type shard_number: int
        @param data: the shard health data
        @type data: dict
        """
        super(Shard, self).__init__()
        self._shard_number = shard_number
        self._data = data

    @property
    def shard_number(self):
        """
        Get the index / number of this shard
        @return:
        @rtype: int
        """
        return self._shard_number

    @property
    def status(self):
        """
        Get the status of this shard
        @return: green, yellow, or red
        @rtype: str
        """
        return self._data['status']

    def is_primary_active(self):
        """
        Is the primary active?
        @return:
        @rtype: bool
        """
        return bool(self._data['primary_active'])

    def is_active(self):
        """
        Is the shard active?
        @return:
        @rtype: bool
        """
        return bool(self._data['active_shards'])

    def is_relocating(self):
        """
        Is the shard relocating?
        @return:
        @rtype: bool
        """
        return bool(self._data['relocating_shards'])

    def is_initialized(self):
        """
        Is the shard initialized?
        @return:
        @rtype: bool
        """
        return bool(self._data['initializing_shards'])

    def is_unassigned(self):
        """
        Is the shard unassigned?
        @return:
        @rtype: bool
        """
        return bool(self._data['unassigned_shards'])



from setuptools import setup, find_packages

long_description = open('README.rst').read()

packages = find_packages(exclude=["tests", "tests.*"])

setup(
    name='pylastica',
    version='1.0.0.0',
    packages=packages,
    url='https://github.com/jlinn/pylastica',
    license='LICENSE.txt',
    author='Joe Linn',
    author_email='',
    description="Python port of Nicolas Ruflin's Elastica PHP library",
    long_description=long_description,
    install_requires=['dateutils', 'urllib3'],
    classifiers=[
        'Intended Audience :: Developers',
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search'
    ]
)



__author__ = 'Joe Linn'

import pylastica.client

class Settings(object):
    def __init__(self, client):
        """
        @param client:
        @type client: pylastica.client.Client
        """
        super(Settings, self).__init__()
        assert isinstance(client, pylastica.client.Client), "client must be an instance of pylastica.client.Client: %r" % client
        self._client = client

    def get(self):
        """
        Returns settings data
        @return: settings data (persistent and transient)
        @rtype: dict
        """
        return self.request().data

    def get_persistent(self, setting=''):
        """
        Returns the current persistent settings of the cluster
        @param setting: Optional. If set, only the specified setting is returned
        @type setting: str
        @return:
        @rtype: dict or str or None
        """
        settings = self.get()['persistent']
        if setting != '':
            if setting in settings:
                return settings[setting]
            else:
                return None
        return settings

    def get_transient(self, setting=''):
        """
        Returns the current transient settings of the cluster
        @param setting: Optional. If set, only returns the specified setting.
        @type setting: str
        @return:
        @rtype: dict or string or None
        """
        settings = self.get()['transient']
        if setting != '':
            if setting in settings:
                return settings[setting]
            else:
                return None
        return settings

    def set_persistent(self, key, value):
        """
        Set a persistent setting
        @param key:
        @type key: str
        @param value:
        @type value: str
        @return:
        @rtype: pylastica.response.Response
        """
        return self.set({'persistent': {key: value}})

    def set_transient(self, key, value):
        """
        Set a transient setting
        @param key:
        @type key: str
        @param value:
        @type value: str
        @return:
        @rtype: pylastica.response.Response
        """
        return self.set({'transient': {key: value}})

    def set_read_only(self, read_only=True, persistent=False):
        """
        Set the cluster to read-only
        @param read_only:
        @type read_only: bool
        @param persistent: If true, the setting is persistent
        @type persistent: bool
        @return:
        @rtype: pylastica.response.Response
        """
        key = 'cluster.blocks.read_only'
        if persistent:
            return self.set_persistent(key, read_only)
        else:
            return self.set_transient(key, read_only)

    def set(self, settings):
        """
        Set settings for the cluster
        @param settings: raw settings (including persistent or transient)
        @type settings: dict
        @return:
        @rtype: pylastica.response.Response
        """
        return self.request(settings, pylastica.request.Request.PUT)

    @property
    def client(self):
        """
        Get the client object
        @return:
        @rtype: pylastica.client.Client
        """
        return self._client

    def request(self, data=None, method=pylastica.request.Request.GET):
        """
        Sends a settings request
        @param data: optional data dict
        @type data: dict
        @param method: optional HTTP method
        @type method: str
        @return:
        @rtype: pylastica.response.Response
        """
        return self.client.request('_cluster/settings', method, data)



__author__ = 'Joe Linn'

from pylastica.query.abstract import AbstractQuery


class SimpleQueryString(AbstractQuery):
    """
    @see http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/query-dsl-simple-query-string-query.html
    """
    OPERATOR_AND = "and"
    OPERATOR_OR = "or"

    def __init__(self, query, fields=None):
        """
        @param query:
        @type query: str
        @param fields:
        @type fields: list of str
        """
        super(SimpleQueryString, self).__init__()
        self.set_query(query)
        if fields is not None:
            self.set_fields(fields)

    def set_query(self, query):
        """
        Set the querystring for this query
        @param query: see linked documentation for querystring syntax
        @type query: str
        @return:
        @rtype: self
        """
        return self.set_param("query", query)

    def set_fields(self, fields):
        """

        @param fields: the fields on which to perform this query. Defaults to index.query.default_field.
        @type fields: list of str
        @return:
        @rtype: self
        """
        return self.set_param("fields", fields)

    def set_default_operator(self, operator):
        """
        Set the default operator to use if no explicit operator is defined in the query string
        @param operator: see OPERATOR_* constants for options
        @type operator: str
        @return:
        @rtype: self
        """
        return self.set_param("default_operator", operator)

    def set_analyzer(self, analyzer):
        """
        Set the analyzer used to analyze each term of the query
        @param analyzer:
        @type analyzer: str
        @return:
        @rtype: self
        """
        return self.set_param("analyzer", analyzer)


__author__ = 'Joe Linn'

from .abstractfilter import AbstractFilter
import pylastica.script


class Script(AbstractFilter):
    def __init__(self, script=None):
        """
        @param script: optional
        @type script: str or pylastica.script.Script
        """
        super(Script, self).__init__()
        if script is not None:
            self.set_script(script)

    def set_script(self, script):
        """
        Set the script object
        @param script:
        @type script: str or dict or pylastica.script.Script
        @return:
        @rtype: self
        """
        self.params = pylastica.script.Script.create(script).to_dict()
        return self



__author__ = 'Joe Linn'

import abc
import pylastica.param
import pylastica.exception


class AbstractFilter(pylastica.param.Param):
    __metaclass__ = abc.ABCMeta

    def set_cached(self, cached=True):
        """
        Sets the filter cache
        @param cached:
        @type cached: bool
        @return:
        @rtype: self
        """
        return self.set_param('_cache', bool(cached))

    def set_cache_key(self, cache_key):
        """
        Set the filter cache key
        @param cache_key:
        @type cache_key: str
        @return:
        @rtype: self
        """
        self.set_param('_cache_key', str(cache_key))

    def set_name(self, name):
        """
        Set the filter name
        @param name:
        @type name: str
        @return:
        @rtype: self
        """
        return self.set_param('_name', name)


class AbstractGeoDistance(AbstractFilter):
    __metaclass__ = abc.ABCMeta

    LOCATION_TYPE_GEOHASH = 'geohash'
    LOCATION_TYPE_LATLON = 'latlon'

    def __init__(self, key, location):
        """

        @param key:
        @type key: str
        @param location: location as dict or geohash {'lat':40.3, 'lon':45.2}
        @type location: dict or str
        """
        super(AbstractGeoDistance, self).__init__()
        self._location_type = None
        self._key = None
        self._latitude = None
        self._longitude = None
        self._geohash = None
        self.set_key(key)
        self.set_location(location)

    def set_key(self, key):
        """
        Set the key
        @param key:
        @type key: str
        @return:
        @rtype: self
        """
        self._key = str(key)
        return self

    def set_location(self, location):
        """
        Set the location
        @param location:
        @type location: dict or str
        @return:
        @rtype: self
        """
        if isinstance(location, dict):
            if 'lat' in location:
                self.set_latitude(location['lat'])
            else:
                raise pylastica.exception.InvalidException("location['lat'] must be set.")
            if 'lon' in location:
                self.set_longitude(location['lon'])
            else:
                raise pylastica.exception.InvalidException("location['lon'] must be set.")
        elif isinstance(location, str):
            self.set_geohash(location)
        else:
            raise pylastica.exception.InvalidException("location must ba a dict or a string.")
        return self

    def set_latitude(self, lat):
        """

        @param lat:
        @type lat: float
        @return:
        @rtype: self
        """
        self._latitude = float(lat)
        self._location_type = self.LOCATION_TYPE_LATLON
        return self

    def set_longitude(self, lon):
        """

        @param lon:
        @type lon: float
        @return:
        @rtype: self
        """
        self._longitude = float(lon)
        self._location_type = self.LOCATION_TYPE_LATLON
        return self

    def set_geohash(self, geohash):
        """

        @param geohash:
        @type geohash: str
        @return:
        @rtype: self
        """
        self._geohash = geohash
        self._location_type = self.LOCATION_TYPE_GEOHASH
        return self

    def _get_location_data(self):
        """

        @rtype: dict or str
        """
        if self._location_type == self.LOCATION_TYPE_LATLON:
            location = {}
            if self._latitude is not None:
                location['lat'] = self._latitude
            else:
                raise pylastica.exception.InvalidException("Latitude must be set.")
            if self._longitude is not None:
                location['lon'] = self._longitude
            else:
                raise pylastica.exception.InvalidException("Longitude must be set.")
        elif self._location_type == self.LOCATION_TYPE_GEOHASH:
            location = self._geohash
        else:
            raise pylastica.exception.InvalidException("Invalid location type.")
        return location

    def to_dict(self):
        self.set_param(self._key, self._get_location_data())
        return super(AbstractGeoDistance, self).to_dict()


class AbstractMulti(AbstractFilter):
    def __init__(self):
        super(AbstractMulti, self).__init__()
        self._filters = []

    def add_filter(self, filter_object):
        """
        Add a filter
        @param filter_object:
        @type filter_object: AbstractFilter
        @return:
        @rtype: self
        """
        assert isinstance(filter_object, AbstractFilter), "filter must be an instance of an implementation of AbstractFilter: %r" % filter_object
        self._filters.append(filter_object.to_dict())
        return self

    def set_filters(self, filters):
        """
        Set filters
        @param filters:
        @type filters: list of AbstractFilter
        @return:
        @rtype: self
        """
        self._filters = []
        for filter_object in filters:
            self.add_filter(filter_object)
        return self

    def to_dict(self):
        data = super(AbstractMulti, self).to_dict()
        name = self._get_base_name()
        filter_data = data[name]
        if filter_data is None:
            filter_data = self._filters
        else:
            filter_data['filters'] = self._filters
        data[name] = filter_data
        return data



__author__ = 'Joe Linn'

import abc
import copy
import pylastica.param
import pylastica.util


class AbstractTransport(pylastica.param.Param):
    __metaclass__ = abc.ABCMeta

    def __init__(self, connection=None):
        """
        @param connection:
        @type connection:  pylastica.connection.Connection
        """
        super(AbstractTransport, self).__init__()
        self._connection = None
        if connection is not None:
            self.connection = connection

    @property
    def connection(self):
        """

        @return:
        @rtype: pylastica.connection.Connection
        """
        return self._connection

    @connection.setter
    def connection(self, connection):
        """

        @param connection:
        @type connection: pylastica.connection.Connection
        """
        assert isinstance(connection, pylastica.connection.Connection), "connection must be a Connection object: %r" % connection
        self._connection = connection

    @abc.abstractmethod
    def execute(self, request, params):
        """
        Executes a transport request
        @param request: request object
        @type request: pylastica.request.Request
        @param params: hostname, port, path, etc.
        @type params: dict
        @return:
        @rtype: pylastica.response.Response
        """
        pass

    @classmethod
    def create(cls, transport, connection, params=None):
        """
        Create a transport
        @param cls:
        @type cls:
        @param transport: can be the name of a transport ('Http', 'Memcache', etc.), a transport object, or a dict
        @type transport: str or AbstractTransport or dict
        @param connection:
        @type connection: pylastica.connection.Connection
        @param params: parameters for the transport class
        @type params: dict
        @return:
        @rtype: cls
        """
        if isinstance(transport, dict) and 'type' in transport:
            transport_params = copy.deepcopy(transport)
            del transport_params['type']
            if params is None:
                params = {}
            params.update(transport_params)
            transport = transport['type']
        if isinstance(transport, str):
            class_name = 'pylastica.transport.%s' % transport
            transport = pylastica.util.get_class(class_name)()
        if isinstance(transport, AbstractTransport):
            transport.connection = connection
            if params:
                for key, value in params.iteritems():
                    transport.set_param(key, value)
        else:
            raise pylastica.exception.InvalidException("Invalid transport.")
        return transport




__author__ = 'Joe Linn'

from .abstract import AbstractQuery


class Simple(AbstractQuery):
    def __init__(self, query):
        """
        Constructs a query from a dict
        @param query:
        @type query: dict
        """
        super(Simple, self).__init__()
        self.set_query(query)

    def set_query(self, query):
        """
        Set a new query dict
        @param query:
        @type query: dict
        @return:
        @rtype: self
        """
        self._query = query
        return self

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        return self._query



__author__ = 'Joe Linn'


class Info(object):
    def __init__(self, node, params=None):
        """
        @param node:
        @type node: pylastica.node.Node
        @param params: list of parameters to return. (settings, os, process, jvm, thread_pool, network, transport, http)
        @type params: list of str
        """
        super(Info, self).__init__()
        import pylastica.node
        assert isinstance(node, pylastica.node.Node), "node must be of type pylastica.node.Node: %r" % node
        self._node = node
        self.refresh(params)

    def get(self, *args):
        """
        Returns an entry in the data array based on the args.
        Example: get('os', 'mem', 'total') returns total memory of the node's system
        Example: get('os', 'mem') returns a dict with all memory info for the node
        @param args:
        @type args: str
        @return: returns None if the requested data cannot be found
        @rtype: str or dict or None
        """
        data = self.data
        for arg in args:
            if arg in data:
                data = data[arg]
            else:
                return None
        return data

    @property
    def port(self):
        """
        Return the port of the node
        @return:
        @rtype: int
        """
        return self.get('http_address')[6:-7].split(':')[1]

    @property
    def ip(self):
        """
        Get the ip address of the node
        @return:
        @rtype: str
        """
        return self.get('http_address')[6:-7].split(':')[0]

    @property
    def data(self):
        """
        Return all info data
        @return:
        @rtype: dict
        """
        return self._data

    @property
    def node(self):
        """
        Return the Node object
        @return:
        @rtype: pylastica.node.Node
        """
        return self._node

    @property
    def response(self):
        """
        Return the Response object
        @return:
        @rtype: pylastica.response.Response
        """
        return self._response

    @property
    def plugins(self):
        """
        Return a list of the plugins installed on this node
        @return:
        @rtype: list of dict
        """
        if self._params is None or 'plugin' not in self._params:
            #plugins were not retrieved when refresh() was called last. Get them now.
            if isinstance(self._params, list):
                self._params.append('plugin')
            else:
                self._params = ['plugin']
            self.refresh(self._params)
        return self.get('plugins')

    def has_plugin(self, name):
        """
        Determine if the given plugin is installed on this node
        @param name: name of the plugin
        @type name: str
        @return:
        @rtype: bool
        """
        for plugin in self.plugins:
            if plugin['name'] == name:
                return True
        return False

    def refresh(self, params=None):
        """
        Reload all node information. Must be called if said information has changed
        @param params: params to return (defaults to none). (settings, os, process, jvm, thread_pool, network, transport, http)
        @type params: list of str
        @return:
        @rtype: pylastica.response.Response
        """
        self._params = params
        query = {param: True for param in params} if params is not None else None
        self._response = self.node.client.request("_nodes/%s" % self.node.name, query=query)
        data = self.response.data
        self._data = data['nodes'].items()[0][1]



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractFilter

class DocType(AbstractFilter):
    def __init__(self, type_name=None):
        """
        @param type_name: document type name
        @type type_name: str
        """
        super(DocType, self).__init__()
        if type_name is not None:
            self.set_doc_type(type_name)


    def set_doc_type(self, type_name):
        """
        Set the document type
        @param type_name:
        @type type_name: str
        @return:
        @rtype: self
        """
        self._type = type_name
        return self

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        return {
            'type': {'value': self._type}
        }




__author__ = 'Joe Linn'

from .abstract import AbstractQuery
import pylastica.filter.abstractfilter

class Filtered(AbstractQuery):
    def __init__(self, query, filter):
        """

        @param query:
        @type query: pylastica.query.AbstractQuery
        @param filter:
        @type filter: pylastica.filter.AbstractFilter
        """
        super(Filtered, self).__init__()
        self.query = query
        self.filter = filter

    @property
    def query(self):
        """

        @return:
        @rtype: pylastica.query.AbstractQuery
        """
        return self._query

    @query.setter
    def query(self, query):
        """

        @param query:
        @type query: pylastica.query.AbstractQuery
        """
        assert isinstance(query, AbstractQuery), "query must be instance of implementation of AbstractQuery: %r" % query
        self._query = query

    @property
    def filter(self):
        """

        @return:
        @rtype: pylastica.filter.AbstractFilter
        """
        return self._filter

    @filter.setter
    def filter(self, filter):
        """

        @param filter:
        @type filter: pylastica.filter.AbstractFilter
        """
        assert isinstance(filter, pylastica.filter.abstractfilter.AbstractFilter), "filter must be instance of implementation of AbstractFilter: %r" % filter
        self._filter = filter

    def to_dict(self):
        """

        @return:
        @rtype: dict
        """
        return {
            'filtered':{
                'query': self.query.to_dict(),
                'filter': self.filter.to_dict()
            }
        }




__author__ = 'Joe Linn'

import abc
import pylastica.param
import pylastica.script


class AbstractAggregation(pylastica.param.Param):
    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super(AbstractAggregation, self).__init__()
        self._name = None
        self.name = name
        self._aggs = {}

    @property
    def name(self):
        """
        Get the name of this facet
        @return:
        @rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Set the name of this facet.
        @param name:
        @type name: str
        """
        if not isinstance(name, str) or name == '':
            raise TypeError("name must be a string: %r" % name)
        self._name = name

    def add_aggregation(self, aggregation):
        """
        Add a sub-aggregation
        @param aggregation: a sub-aggregation
        @type aggregation: AbstractAggregation
        @return:
        @rtype: self
        """
        if not isinstance(aggregation, AbstractAggregation):
            raise TypeError("aggregation must be an implementation of AbstractAggregation: %r" % aggregation)
        self._aggs[aggregation.name] = aggregation.to_dict()
        return self

    @property
    def aggs(self):
        return self._aggs

    def to_dict(self):
        """
        @return:
        @rtype: dict
        """
        dictionary = super(AbstractAggregation, self).to_dict()
        if len(self._aggs):
            dictionary['aggs'] = self._aggs
        return dictionary


class SimpleAggregation(AbstractAggregation):
    __metaclass__ = abc.ABCMeta

    def set_field(self, field):
        """
        Set the field for this aggregation
        @param field: the name of the document field on which to perform this aggregation
        @type field: str
        @return:
        @rtype: self
        """
        return self.set_param("field", field)

    def set_script(self, script):
        """
        Set a script for this aggregation
        @param script: if a Script object is passed, its params will be used
        @type script: str or Script
        @return:
        @rtype: self
        """
        if isinstance(script, pylastica.script.Script):
            self.set_param("params", script.params)
            script = script.script
        return self.set_param("script", script)


__author__ = 'Joe Linn'

from .index import *
from .shard import *

class Health(object):
    def __init__(self, client):
        """
        @param client:
        @type client: pylastica.client.Client
        """
        super(Health, self).__init__()
        self._client = client
        self.refresh()
        self._data = None

    def _retrieve_health_data(self):
        """
        Retrieves health data from the cluster
        @return:
        @rtype: dict
        """
        return self._client.request('_cluster/health', query={'level': 'shards'}).data

    @property
    def data(self):
        """
        Get the health data
        @return:
        @rtype: dict
        """
        return self._data

    def refresh(self):
        """
        Refresh the health data of the cluster
        @return:
        @rtype: self
        """
        self._data = self._retrieve_health_data()
        return self

    @property
    def cluster_name(self):
        """
        Get the name of the cluster
        @return:
        @rtype: str
        """
        return self._data['cluster_name']

    @property
    def status(self):
        """
        Get the status of the cluster
        @return: green, yellow, or red
        @rtype: str
        """
        return self._data['status']

    @property
    def timed_out(self):
        """

        @return:
        @rtype: bool
        """
        return bool(self._data['timed_out'])

    @property
    def number_of_nodes(self):
        """
        Get the number of nodes in the cluster
        @return:
        @rtype: int
        """
        return int(self._data['number_of_nodes'])

    @property
    def number_of_data_nodes(self):
        """
        Get the number of data nodes in the cluster
        @return:
        @rtype: int
        """
        return int(self._data['number_of_data_nodes'])

    @property
    def active_primary_shards(self):
        """
        Get the number of active primary shards
        @return:
        @rtype: int
        """
        return int(self._data['active_primary_shards'])

    @property
    def active_shards(self):
        """
        Get the number of active shards
        @return:
        @rtype: int
        """
        return int(self._data['active_shards'])

    @property
    def relocating_shards(self):
        """
        Get the number of relocating shards
        @return:
        @rtype: int
        """
        return int(self._data['relocating_shards'])

    @property
    def initializing_shards(self):
        """
        Get the number of initializing shards
        @return:
        @rtype: int
        """
        return int(self._data['initializing_shards'])

    @property
    def unassigned_shards(self):
        """
        Get the number of unassigned shards
        @return:
        @rtype: int
        """
        return int(self._data['unassigned_shards'])

    @property
    def inidices(self):
        """
        Get the status of the indices
        @return:
        @rtype: pylastica.cluster.health.index.Index
        """
        return [Index(name, index) for name, index in self._data['indices'].iteritems()]



__author__ = 'Joe Linn'

#import pylastica.filter.abstractfilter
from .abstractfilter import AbstractMulti

class BoolOr(AbstractMulti):
    def _get_base_name(self):
        """

        @return:
        @rtype: str
        """
        return 'or'



__author__ = 'Joe Linn'

import urllib3
import json
import pylastica.transport
import pylastica.exception.connection


class Http(pylastica.transport.AbstractTransport):
    _scheme = 'http'

    METHOD_GET = 'GET'
    METHOD_POST = 'POST'
    METHOD_PUT = 'PUT'
    METHOD_DELETE = 'DELETE'

    def __init__(self, connection=None):
        super(Http, self).__init__(connection)
        self._headers = {}

    def execute(self, request, params):
        """
        Executes a transport request
        @param request: request object
        @type request: pylastica.request.Request
        @param params: hostname, port, path, etc.
        @type params: dict
        @return:
        @rtype: pylastica.response.Response
        """
        connection = self.connection
        url = connection.get_config('url') if connection.has_config('url') else ''
        if url != '' and url is not None:
            base_url = url
        else:
            connection_data = {
                'scheme': self._scheme,
                'host': connection.host,
                'port': connection.port,
                'path': connection.path
            }
            base_url = "%(scheme)s://%(host)s:%(port)s/%(path)s" % connection_data
        base_url += request.path
        query = request.query #http query params
        if connection.has_config('headers'):
            for key, value in connection.get_config('headers'):
                self.setHeader(key, value)
        data = request.get_data()
        http_method = request.get_method()
        content = None
        if data is not None:
            if self.has_param('postWithRequestBody') and self.get_param('postWithRequestBody'):
                http_method = self.METHOD_POST
            if isinstance(data, dict) or isinstance(data, list):
                content = json.dumps(data)
            else:
                content = data
        #TODO: something for debugging
        response = self.request(http_method, base_url, query, content)
        response_object = pylastica.response.Response(response.data, response.status)
        return response_object

    def setHeader(self, key, value):
        """
        Set a HTTP header. Will overwrite any existing value with the same key.
        @param key: The key of the header
        @type key: str
        @param value: The value of the header
        @type value: str
        @return: self
        @rtype: Http
        """
        self._headers[key] = value
        return self

    def clearHeaders(self):
        """
        Clear currently set header data
        @return:
        @rtype: void
        """
        self._headers = {}

    def get(self, url, query_params={}, data=None):
        """

        @param url: full-qualified URL
        @param query_params: any query parameters to be sent along with the URL
        @param data: request data to be sent
        @return: results of the HTTP GET request
        """
        return self.request(self.METHOD_GET, url, query_params, data)

    def post(self, url, query_params={}, data=None):
        """

        @param url:
        @type url: str
        @param query_params:
        @type query_params: dict
        @param data:
        @type data: str
        @return:
        @rtype: str
        """
        return self.request(self.METHOD_POST, url, query_params, data)

    def put(self, url, query_params={}, data=None):
        """

        @param url:
        @type url: str
        @param query_params:
        @type query_params: dict
        @param data:
        @type data: str
        @return: str
        @rtype:
        """
        return self.request(self.METHOD_PUT, url, query_params, data)

    def delete(self, url, query_params={}, data=None):
        """

        @param url:
        @type url: str
        @param query_params:
        @type query_params: dict
        @param data:
        @type data: str
        @return:
        @rtype: str
        """
        return self.request(self.METHOD_DELETE, url, query_params, data)

    def request(self, method, url, query_params=None, data=None):
        """
        @param method: HTTP method (see METHOD_* constants)
        @type method: str
        @param url:
        @type url: str
        @param query_params:
        @type query_params: dict
        @param data:
        @type data: str
        @return:
        @rtype: urllib3.response.HTTPResponse
        """
        http = urllib3.PoolManager()
        try:
            response = http.request_encode_url(method, url, fields=query_params, body=data, headers=self._headers)
        except urllib3.exceptions.MaxRetryError as e:
            raise pylastica.exception.connection.HttpException(e)
        return response



__author__ = 'Joe Linn'

from .abstract import AbstractQuery
import pylastica.doc_type


class Ids(AbstractQuery):
    def __init__(self, doc_type=None, ids=None):
        """
        @param doc_type: document type
        @type doc_type: str or pylastica.doc_type.DocType
        @param ids: document ids
        @type ids: list of str
        """
        super(Ids, self).__init__()
        if doc_type is not None:
            self.set_type(doc_type)
        if ids is not None:
            self.set_ids(ids)

    def add_id(self, doc_id):
        """
        Add another id to the query
        @param doc_id:
        @type doc_id: str
        @return:
        @rtype: self
        """
        if 'values' not in self._params or self._params['values'] is None:
            self._params['values'] = []
        self._params['values'].append(doc_id)
        return self

    def add_type(self, doc_type):
        """
        Add a document type to the query
        @param doc_type:
        @type doc_type: str or pylastica.doc_type.DocType
        @return:
        @rtype: self
        """
        if isinstance(doc_type, pylastica.doc_type.DocType):
            doc_type = doc_type.name
        elif doc_type == '':
            raise pylastica.exception.InvalidException("Document types cannot be empty strings.")
        if 'type' not in self._params or self._params['type'] is None:
            self._params['type'] = []
        self._params['type'].append(doc_type)
        return self

    def set_type(self, doc_type):
        """
        Set the doc type for this query. Overwrites all current values.
        @param doc_type:
        @type doc_type: str or pylastica.doc_type.DocType
        @return:
        @rtype: self
        """
        if isinstance(doc_type, pylastica.doc_type.DocType):
            doc_type = doc_type.name
        elif doc_type == '':
            raise pylastica.exception.InvalidException("Document types cannot be empty strings.")
        self._params['type'] = [doc_type]
        return self

    def set_ids(self, ids):
        """
        Set document ids. Overwrites all current values.
        @param ids:
        @type ids: list of str
        @return:
        @rtype: self
        """
        assert isinstance(ids, list), "ids must be a list: %r" % ids
        self._params['values'] = ids
        return self




__author__ = 'Joe Linn'

from .abstract import AbstractQuery

class Prefix(AbstractQuery):
    def __init__(self, prefix=None):
        """

        @param prefix:
        @type prefix: dict of str
        """
        super(Prefix, self).__init__()
        self.set_raw_prefix(prefix)

    def set_raw_prefix(self, prefix):
        """
        Set values for the prefix
        @param prefix:
        @type prefix: dict
        @return:
        @rtype: self
        """
        if prefix is None:
            prefix = {}
        self.params = prefix
        return self

    def set_prefix(self, key, value, boost=1.0):
        """
        Set the prefix for this query
        @param key: key to query
        @type key: str
        @param value: value(s) for the query
        @type value: str or list of str
        @param boost: optional, defaults to 1.0
        @type boost: float
        @return:
        @rtype: self
        """
        return self.set_raw_prefix({key: {'value': value, 'boost': boost}})



__author__ = 'Joe Linn'

from .datehistogram import *
from .filter import *
from .geodistance import *
from .histogram import *
from .query import *
from .range import *
from .statistical import *
from .terms import *
from .termsstats import *



__author__ = 'Joe Linn'

from pylastica.param import Param


class Suggest(Param):
    """
    @see: http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/search-suggesters.html
    """
    def __init__(self, suggestion=None):
        """
        @param suggestion: optional
        @type suggestion:  pylastica.suggest.abstract.AbstractSuggestion
        """
        super(Suggest, self).__init__()
        if suggestion is not None:
            self.add_suggestion(suggestion)

    def set_global_text(self, text):
        """
        Set the global text for this suggester
        @param text:
        @type text: str
        @rtype: self
        """
        return self.set_param("text", text)

    def add_suggestion(self, suggestion):
        """
        Add a suggestion to this suggest clause
        @param suggestion:
        @type suggestion: pylastica.suggest.abstract.AbstractSuggestion
        @return:
        @rtype: self
        """
        self.set_param(suggestion.name, suggestion.to_dict())
